from __future__ import annotations
import base64
from enum import Enum
import inspect
from itertools import zip_longest
import re
import struct
import threading
import typing
from typing import Any, Dict, List, Optional, Set, Tuple, Union, get_type_hints
from relationalai.clients.util import IdentityParser, ParseError
from typing_extensions import TypeGuard
import numbers
import os
import datetime
import hashlib
import traceback
import rich

from pandas import DataFrame

from relationalai.environments import runtime_env, SnowbookEnvironment

from .clients.client import Client

from .metamodel import Behavior, Builtins, ActionType, Var, Task, Action, Builder, Type as mType, Property as mProperty
from . import debugging, errors
from .errors import FilterAsValue, Errors, InvalidPropertySetException, MultipleIdentities, NonCallablePropertyException, OutOfContextException, RAIException, ReservedPropertyException, VariableOutOfContextException, handle_missing_integration

#--------------------------------------------------
# Constants
#--------------------------------------------------

RESERVED_PROPS = ["add", "set", "persist", "unpersist"]

Value = Union[
    "Expression",
    "ContextSelect", # because of union/match
    int,
    float,
    str,
    bool,
    datetime.datetime,
    datetime.date,
]

#--------------------------------------------------
# Helpers
#--------------------------------------------------

# @FIXME: This should only return a Var or None but that's gonna take some doing.
def to_var(x:Any) -> Any:
    if isinstance(x, Var):
        return x
    if getattr(x, "_to_var", None):
        return to_var(x._to_var())
    if isinstance(x, ContextSelect):
        return x._vars[0]
    if isinstance(x, mProperty):
        return Var(value=x)
    if isinstance(x, mType):
        return Var(value=x)
    if isinstance(x, Producer):
        x._use_var()
        return x._var
    if isinstance(x, list) or isinstance(x, tuple):
        return Var(Builtins.Any, value=[v for i in x if (v := to_var(i))])
    if isinstance(x, str):
        return Var(Builtins.String, None, x)
    if isinstance(x, numbers.Number):
        return Var(Builtins.Number, None, x)
    if isinstance(x, datetime.datetime) or isinstance(x, datetime.date):
        return Var(value=x)
    raise Exception(f"Unknown type: {type(x)}\n{x}")

build = Builder(to_var)

def to_list(x:Any):
    if isinstance(x, list):
        return x
    if isinstance(x, tuple):
        return list(x)
    return [x]

def is_static(x:Any):
    if isinstance(x, Var):
        return x.value is not None
    if isinstance(x, Type):
        return True
    if isinstance(x, str):
        return True
    if isinstance(x, Producer):
        return is_static(to_var(x))
    if isinstance(x, numbers.Number):
        return True
    if isinstance(x, list):
        return all(is_static(i) for i in x)
    if isinstance(x, tuple):
        return all(is_static(i) for i in x)
    if isinstance(x, dict):
        return all(is_static(i) for i in x.values())
    return False

def is_collection(x:Any) -> TypeGuard[Union[List, Tuple, Set]]:
    return isinstance(x, list) or isinstance(x, tuple) or isinstance(x, set)

def rel_dict(**kwargs):
    return InlineRelation(get_graph(), [
        (Symbol(k), v) for k, v in kwargs.items()
    ])

#--------------------------------------------------
# Base
#--------------------------------------------------

id = 0
def next_id():
    global id
    id += 1
    return id

#--------------------------------------------------
# Producer
#--------------------------------------------------

attributes_to_skip = {
    "_to_var",
    "__name__",
    "__qualname__",
    "_repr_html_",
    "_supports_binary_op",
}

def is_called_by_streamlit_write():
    stack_frames = inspect.stack()

    for frame in stack_frames[1:]:
        frame_info = frame
        context = frame_info.code_context[0] if frame_info.code_context else ""

        if (
            frame_info.frame.f_globals.get("__name__") == "streamlit.runtime.scriptrunner.magic_funcs"
            and "st.write" in context
        ):
            return True

    return False

class Producer():
    def __init__(self, graph:'Graph'):
        self._id = next_id()
        self._graph = graph
        self._subs = {}
        self._var = None

    def __iter__(self):
        if is_called_by_streamlit_write():
            return iter([])
        raise Exception("Can't iterate over a producer")

    def __repr__(self):
        if isinstance(runtime_env, SnowbookEnvironment):
            # suppress spurious output in Snowflake notebooks
            # for expressions like p.age > 18
            return ""
        else:
            return super().__repr__()

    def __getattr__(self, name: str) -> Any:
        if is_called_by_streamlit_write():
            return None
        if name in attributes_to_skip:
            return None
        if name == "getdoc":
            rich.print("[red bold]GETDOC CALLED")
            traceback.print_stack()
            return None
        self._subs[name] = self._make_sub(name, self._subs.get(name))
        return self._subs[name]

    def _make_sub(self, name:str, existing:Optional['Producer']=None) -> Any:
        raise Exception("Implement Producer._make_sub")

    def _use_var(self):
        pass

    #--------------------------------------------------
    # Boolean overloads
    #--------------------------------------------------

    def __bool__(self):
        # This doesn't seem to be safe as Python can call bool on Producers in lots of random cases
        # InvalidBoolWarning(Errors.call_source())
        # class_name = self.__class__.__name__
        # TODO: Refactor this to be Exception instead of warning, so we could directly raise it if we bring back this
        # raise TypeError(f"Can't convert an {class_name} to a boolean.")
        return True

    #--------------------------------------------------
    # Math overloads
    #--------------------------------------------------

    def _wrapped_op(self, op, left, right):
        # Check if an active context redefines this binary operator (e.g. for the solvers library).
        for s in self._graph._stack.stack:
            f = getattr(s, "_supports_binary_op", None)
            if f and f(op):
                return s._make_binary_op(op, left, right)
        return Expression(self._graph, op, [left, right])

    def __add__(self, other):
        return self._wrapped_op(Builtins.plus, self, other)
    def __radd__(self, other):
        return self._wrapped_op(Builtins.plus, other, self)

    def __mul__(self, other):
        return self._wrapped_op(Builtins.mult, self, other)
    def __rmul__(self, other):
        return self._wrapped_op(Builtins.mult, other, self)

    def __sub__(self, other):
        return self._wrapped_op(Builtins.minus, self, other)
    def __rsub__(self, other):
        return self._wrapped_op(Builtins.minus, other, self)

    def __truediv__(self, other):
        return self._wrapped_op(Builtins.div, self, other)
    def __rtruediv__(self, other):
        return self._wrapped_op(Builtins.div, other, self)

    def __floordiv__(self, other):
        return self._wrapped_op(Builtins.floor_div, self, other)
    def __rfloordiv__(self, other):
        return self._wrapped_op(Builtins.floor_div, other, self)

    def __pow__(self, other):
        return self._wrapped_op(Builtins.pow, self, other)
    def __rpow__(self, other):
        return self._wrapped_op(Builtins.pow, other, self)

    def __mod__(self, other):
        return self._wrapped_op(Builtins.mod, self, other)
    def __rmod__(self, other):
        return self._wrapped_op(Builtins.mod, other, self)

    def __neg__(self):
        return self._wrapped_op(Builtins.mult, self, -1)

    #--------------------------------------------------
    # Filter overloads
    #--------------------------------------------------

    def __gt__(self, other):
        return self._wrapped_op(Builtins.gt, self, other)
    def __ge__(self, other):
        return self._wrapped_op(Builtins.gte, self, other)
    def __lt__(self, other):
        return self._wrapped_op(Builtins.lt, self, other)
    def __le__(self, other):
        return self._wrapped_op(Builtins.lte, self, other)
    def __eq__(self, other) -> Any:
        return self._wrapped_op(Builtins.approx_eq, self, other)
    def __ne__(self, other) -> Any:
        return self._wrapped_op(Builtins.neq, self, other)

    #--------------------------------------------------
    # Context management
    #--------------------------------------------------

    def __enter__(self):
        self._graph._push(self)

    def __exit__(self, *args):
        self._graph._pop(self)

#--------------------------------------------------
# Context
#--------------------------------------------------

class TaskExecType(Enum):
    Query = 1
    Rule = 2
    Procedure = 3
    Export = 4

class ContextSelect(Producer):
    def __init__(self, context:'Context'):
        super().__init__(context.graph)
        self._context = context
        self._select_len = None
        self._insts = []
        self._vars = []
        self._props = {}

    def _assign_vars(self):
        task = self._context._task
        if not len(self._vars) and self._select_len:
            self._insts = to_list(Vars(self._select_len))
            self._vars = [to_var(v) for v in self._insts]
            task.properties = [Builtins.Relation.properties[i] for i in range(self._select_len)]
            task.bindings.update({Builtins.Relation.properties[i]: v for i, v in enumerate(self._vars)})

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            return None
        if name in self._props:
            return Instance(self._context.graph, ActionType.Get, [], {}, var=to_var(self._props[name]))
        else:
            return getattr(Instance(self._context.graph, ActionType.Get, [], {}, var=to_var(self._vars[0])), name)

    #--------------------------------------------------
    # Return handling
    #--------------------------------------------------

    def _do_return(self, args, distinct=False):
        graph = self._context.graph
        task = self._context._task
        if task.behavior == Behavior.Query \
            and self._context._exec_type in [TaskExecType.Query, TaskExecType.Procedure]:
            if isinstance(args[0], tuple):
                args = args[0]
            export = self._context._format == "snowpark"
            graph._action(build.return_(list(args), export=export, distinct=distinct))
        else:
            #TODO: good error message depending on the type of task we're dealing with
            raise Exception("Can't select in a non-query")
        return self._context

    def __call__(self, *args: Any) -> Any:
        if self._context._done:
            raise errors.SelectOutOfContext()
        return self._do_return(args)

    def distinct(self, *args:Any) -> Any:
        if self._context._done:
            raise errors.SelectOutOfContext()
        return self._do_return(args, distinct=True)

    #--------------------------------------------------
    # Add for `model.match() as m`
    #--------------------------------------------------

    def add(self, item, **kwargs):
        arg_len = len(kwargs) + 1
        if self._select_len is not None and arg_len != self._select_len:
            raise Exception("Add must be provided the same arguments in each branch")
        self._select_len = arg_len
        self._assign_vars()
        if len(self._props) and set(self._props.keys()) != set(kwargs.keys()):
            raise Exception("Add must be provided the same properties in each branch")
        elif len(self._props) == 0:
            for k, v in zip(kwargs.keys(), self._vars[1:]):
                v.name = k
                self._props[k] = v

        graph = self._context.graph
        graph._action(build.relation_action(ActionType.Bind, self._context._task, [item, *[kwargs[k] for k in self._props.keys()]]))

class Context():
    def __init__(self, graph:'Graph', *args, behavior=Behavior.Query, op=None,
                 exec_type=TaskExecType.Rule, dynamic=False, format="default",
                 props=None, engine=None, tag=None, globalize=False, source=None):
        self._id = next_id()
        self.results = DataFrame()
        self.graph = graph
        self._task = Task(behavior=behavior)
        self._globalize = globalize
        self._op = op
        self._args = list(args)
        self._exec_type = exec_type
        self._select_len = None
        self._dynamic = dynamic or any([item._dynamic for item in self.graph._stack.stack if isinstance(item, Context)])
        self._props = props or {}
        self._engine= engine
        self._tag = tag # for benchmark reporting
        self._format = self._resolve_format(format)
        self._done = False
        self._source = source

    def _resolve_format(self, format):
        if format == "default":
            if self.graph._format != "default":
                return self.graph._format
            if self.graph._config.get("platform") == "snowflake":
                # In the future we may want to default to snowpark
                return "pandas"
            return "pandas"
        return format

    def __enter__(self):
        debugging.set_source(self._task, self._source)
        self.graph._push(self)
        return ContextSelect(self)

    def __exit__(self, *args):
        self._done = True
        # If no exception info has been passed to args,
        # then proceed with the normal exit process.
        # Otherwise, return False to propagate the exception.
        if args[0] is None:
            if not self._dynamic:
                debugging.check_errors(self._task)
            try:
                self.graph._pop(self, globalize=self._globalize)
            except KeyboardInterrupt as e:
                print("Canceling transactions...")
                self.graph.resources.cancel_pending_transactions()
                raise e
            except RAIException as e:
                handle_missing_integration(e)
                raise e.clone() from None
        else:
            self.graph._pop(self, exec=False)
            if isinstance(args[1], RAIException):
                raise args[1].clone() from None
        return False

    def __iter__(self):
        if self._exec_type != TaskExecType.Query:
            raise Exception("Can't iterate over a non-query task")
        else:
            return self.results.itertuples(index=False)

    def _repr_html_(self):
        if self._exec_type == TaskExecType.Query:
            if isinstance(self.results, DataFrame):
                return self.results.to_html(index=False)
            elif getattr(self.results, "_repr_html_", None):
                return self.results._repr_html_()
            elif getattr(self.results, "show"):
                self.results.show()
                return ""
            else:
                return str(self.results)

    def __str__(self):
        if self._exec_type == TaskExecType.Query:
            if isinstance(self.results, DataFrame):
                return self.results.to_string(index=False)
            else:
                return str(self.results)
        return super().__str__()

#--------------------------------------------------
# Type
#--------------------------------------------------

def hash_values_sha256_truncated(args):
    combined = ''.join(map(str, args))
    combined_bytes = combined.encode('utf-8')
    hasher = hashlib.sha256()
    hasher.update(combined_bytes)
    hash_128_bit = hasher.digest()[:16]
    return hash_128_bit

# @NOTE: `omit_intrinsic_type_in_hash` exists to keep node hashes (which are namespaced by their graph) stable
#        when new nodes are directly created via `Node.add(x = ..)` instead of marking existing entities as
#        nodes. Possible hash collisions between types happen to be okay in this case because of the access
#        paths, but it's _not_ generally safe to use it otherwise.
class Type(Producer):
    def __init__(self, graph:'Graph', name:str, builtins:List[str] = [], scope:str="", omit_intrinsic_type_in_hash = False):
        super().__init__(graph)
        self._type = mType(scope+name)
        self._scope = scope
        self._omit_intrinsic_type_in_hash = omit_intrinsic_type_in_hash
        if graph._config.get("compiler.use_value_types", False):
            self._type.parents.append(Builtins.ValueType)
        install = build.install(self._type)
        self._graph._action(install)
        debugging.set_source(install)

    def __call__(self, *args, **kwargs):
        return Instance(self._graph, ActionType.Get, [self, *args], kwargs, name=self._type.name.lower(), scope=self._scope)

    def add(self, *args, **kwargs):
        inst = Instance(self._graph, ActionType.Bind, [self, *args], kwargs, name=self._type.name.lower(), is_add=True, scope=self._scope)
        if inst._action.entity.value is not None:
            pass
        elif is_static(args) and is_static(kwargs):
            params = [Var(value=t.name) for t in inst._action.types]
            if self._omit_intrinsic_type_in_hash:
                params.pop(0)
            params.extend(inst._action.bindings.values())
            inst._action.entity.value = hash_values_sha256_truncated(params)
        elif all([isinstance(a, Type) for a in args]):
            self._graph._action(build.ident(inst._action, self._omit_intrinsic_type_in_hash))
        inst._add_to_graph()
        return inst

    def persist(self, *args, **kwargs):
        inst = Instance(self._graph, ActionType.Persist, [self, *args], kwargs, name=self._type.name.lower(), is_add=True, scope=self._scope)
        if inst._action.entity.value is not None:
            pass
        elif is_static(args) and is_static(kwargs):
            params = [Var(value=t.name) for t in inst._action.types]
            params.extend(inst._action.bindings.values())
            inst._action.entity.value = hash_values_sha256_truncated(params)
        elif all([isinstance(a, Type) for a in args]):
            self._graph._action(build.ident(inst._action, self._omit_intrinsic_type_in_hash))
        inst._add_to_graph()
        return inst

    def extend(self, *args, **kwargs):
        for arg in args:
            if not isinstance(arg, Type):
                raise Exception("Can only extend a type with another type")
            with self._graph.rule(dynamic=True):
                a = arg()
                a.set(self)
            with self._graph.rule(dynamic=True):
                a = arg()
                neue = self(a)
                for k, v in kwargs.items():
                    if isinstance(v, Property):
                        v = getattr(a, v._prop.name)
                    neue.set(**{k:v})

    def define(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, tuple):
                (other_type, left, right) = v
                with self._graph.rule():
                    inst = other_type()
                    me = self()
                    getattr(inst, right) == getattr(me, left) # type: ignore
                    if getattr(self, k).is_multi_valued:
                        getattr(me, k).add(inst)
                    else:
                        me.set(**{k: inst})
            else:
                raise Exception("Define requires a tuple of (Type, left, right)")
        return self

    def keyed(self, **kwargs):
        return KeyedType(self, kwargs)

    def __or__(self, __value: Any) -> TypeUnion:
        if isinstance(__value, Type) or isinstance(__value, TypeIntersection):
            return TypeUnion(self._graph, [self, __value])
        if isinstance(__value, TypeUnion):
            return TypeUnion(self._graph, [self, *__value._types])
        raise Exception("Can't or a type with a non-type")

    def __and__(self, __value: Any) -> TypeIntersection:
        if isinstance(__value, Type) or isinstance(__value, TypeUnion):
            return TypeIntersection(self._graph, [self, __value])
        if isinstance(__value, TypeIntersection):
            return TypeIntersection(self._graph, [self, *__value._types])
        raise Exception("Can't & a type with a non-type")

    def _make_sub(self, name: str, existing=None):
        if existing is not None:
            return existing
        return Property(self._graph, name, [self._type], self, scope=self._scope)

    def known_properties(self):
        return [p.name.removeprefix(self._scope) for p in self._type.properties]

#--------------------------------------------------
# KeyedType
#--------------------------------------------------

class KeyedType(Producer):
    def __init__(self, type:Type, keys:Dict[str,Type]):
        super().__init__(type._graph)
        self._type = type
        self._relation = RawRelation(self._graph, type._type.name, len(keys))
        self._key_info = keys
        self._key_order = list(keys.keys())
        self._props = {}

        # declare the relation
        install = build.install(self._relation._type)
        self._graph._action(install)
        debugging.set_source(install)

    def _prop(self, name:str):
        if name not in self._props:
            self._props[name] = RawRelation(self._graph, f"{self._relation._name}_{name}", len(self._key_info)+1)
        return self._props[name]

    def extend(self, *args, **kwargs):
        raise errors.KeyedCantBeExtended()

    def _collect_keys(self, args:List[Any], kwargs:Dict[str,Any]):
        ks = self._key_info.keys()
        all_args = dict(zip(ks, args))
        all_args.update(kwargs)
        props = set(all_args.keys()) - set(ks)
        return all_args, {k: all_args.get(k) for k in ks if k in all_args}, {k: all_args[k] for k in props}

    def add(self, *args:Any, **kwargs):
        all_args, keys, props = self._collect_keys(list(args), kwargs)
        if len(keys) != len(self._key_info):
            raise errors.KeyedWrongArity(self._type._type.name, list(self._key_info.keys()), len(keys))
        self._relation.add(*keys.values())
        for k, v in props.items():
            self._prop(k).add(*keys, v)
        return KeyedInstance(self, list(keys), {})

    def __call__(self, *args, **kwargs):
        all_args, keys, props = self._collect_keys(list(args), kwargs)
        ks = self._key_info.keys()
        keys = [all_args.get(k, Vars(1)) for k in ks]
        self._relation(*keys)
        return KeyedInstance(self, keys, props)

    def __getattr__(self, name: str) -> Any:
        return self._props[name]

class KeyedInstance(Producer):
    def __init__(self, keyed_type:KeyedType, keys:List[Any], kwargs:dict={}):
        self._keys = keys
        self._type = keyed_type
        for k, v in kwargs.items():
            self._type._prop(k)(*keys, v)

    def __getattr__(self, name:str):
        if name in self._type._key_order:
            return self._keys[self._type._key_order.index(name)]
        v = Vars(1)
        self._type._prop(name)(*self._keys, v)
        return v

    def set(self, **kwargs):
        for k, v in kwargs.items():
            self._type._prop(k).add(*self._keys, v)
        return self

    def __iter__(self):
        return iter(self._keys)

    def _to_var(self):
        raise Exception("KeyedTypes can't be returned directly, you can only reference properties")

#--------------------------------------------------
# TypeUnion
#--------------------------------------------------

class TypeUnion(Producer):
    def __init__(self, graph:'Graph', types:List[Type|TypeIntersection]):
        super().__init__(graph)
        self._types = types

    def __call__(self, *args, **kwargs) -> 'ContextSelect':
        if not len(self._graph._stack.stack):
            raise Exception("Can't create an instance outside of a context")
        graph = self._graph
        with graph.union(dynamic=True) as union:
            for t in self._types:
                with graph.scope():
                    union.add(t(*args, **kwargs))
        return union

    def known_properties(self):
        props = []
        for t in self._types:
            for prop in t.known_properties():
                if prop not in props:
                    props.append(prop)
        return props

    def __or__(self, __value: Any) -> 'TypeUnion':
        if isinstance(__value, Type):
            return TypeUnion(self._graph, [*self._types, __value])
        if isinstance(__value, TypeUnion):
            return TypeUnion(self._graph, [*self._types, *__value._types])
        raise Exception("Can't or a type with a non-type")

    def _make_sub(self, name: str, existing=None):
        if existing is not None:
            return existing
        return Property(self._graph, name, [t._type for t in self._types], self)

#--------------------------------------------------
# TypeIntersection
#--------------------------------------------------

class TypeIntersection(Producer):
    def __init__(self, graph:'Graph', types:List[Type|TypeUnion]):
        super().__init__(graph)
        self._types = types

    def __call__(self, *args, **kwargs) -> Producer:
        if not len(self._graph._stack.stack):
            raise Exception("Can't create an instance outside of a context")
        return self._types[0](*self._types[1:], *args, **kwargs)

    def known_properties(self):
        props = []
        for t in self._types:
            for prop in t.known_properties():
                if prop not in props:
                    props.append(prop)
        return props

    def __and__(self, __value: Any) -> TypeIntersection:
        if isinstance(__value, Type):
            return TypeIntersection(self._graph, [*self._types, __value])
        if isinstance(__value, TypeIntersection):
            return TypeIntersection(self._graph, [*self._types, *__value._types])
        raise Exception("Can't & a type with a non-type")

    def _make_sub(self, name: str, existing=None):
        if existing is not None:
            return existing
        return Property(self._graph, name, [t._type for t in self._types], self)

#--------------------------------------------------
# Property
#--------------------------------------------------

class Property(Producer):
    def __init__(self, graph:'Graph', name:str, types:List[mType], provider:Type|TypeUnion|TypeIntersection, scope:str=""):
        super().__init__(graph)
        self._name = name
        self._type = types[0]
        self._scope = scope
        self._provider = provider
        self._prop = build.property_named(scope+name, types)

    def __call__(self, key:Any, value:Any):
        action = build.relation_action(ActionType.Get, self._prop, [key, value])
        self._graph._action(action)

    def _use_var(self):
        raise Exception("Support properties being used as vars")

    def _make_sub(self, name: str, existing=None):
        raise Exception("Support properties on properties?")

    def to_property(self):
        return self._prop

    def declare(self):
        self._graph._action(build.install(self._prop))

    def has_many(self):
        self.declare()
        self._graph._check_property(self._prop, multi_valued=True)

    @property
    def is_multi_valued(self):
        return self._graph._prop_is_multi.get(self._name)

#--------------------------------------------------
# Instance
#--------------------------------------------------

class Instance(Producer):
    def __init__(self, graph:'Graph', action_type:ActionType, positionals:List[Any], named:Dict[str,Any], var:Var|None=None, name=None, is_add=False, scope:str=""):
        super().__init__(graph)
        self._action = Action(action_type, to_var(var) if var else Var(name=name))
        self._actions = [self._action]
        self._sets = {}
        self._context = graph._stack.active()
        self._scope = scope
        available_types = []
        last_pos_var = None

        #--------------------------------------------------
        # Positionals
        #--------------------------------------------------
        has_ident = False
        for pos in positionals:
            if isinstance(pos, Type):
                self._action.append(pos._type)
            elif isinstance(pos, Instance):
                if has_ident:
                    raise MultipleIdentities()
                has_ident = True
                self._action.append(to_var(pos))
                available_types.extend(pos._action.types)
                if last_pos_var:
                    self._graph._action(build.eq(last_pos_var, self._action.entity))
                last_pos_var = self._action.entity
            elif isinstance(pos, TypeUnion) or isinstance(pos, TypeIntersection):
                self._action.append(to_var(pos()))
                available_types.extend([t._type for t in pos._types])
                if last_pos_var:
                    self._graph._action(build.eq(last_pos_var, self._action.entity))
                last_pos_var = self._action.entity
            elif isinstance(pos, Producer):
                if has_ident:
                    raise MultipleIdentities()
                has_ident = True
                self._action.append(to_var(pos))
                if last_pos_var:
                    self._graph._action(build.eq(last_pos_var, self._action.entity))
                last_pos_var = self._action.entity
            elif isinstance(pos, (str, int)):
                if has_ident:
                    raise MultipleIdentities()
                has_ident = True
                if isinstance(pos, str):
                    try:
                        decoded = base64.b64decode(pos + "==")
                        if len(decoded) == 16:
                            first_int = decoded[:8]
                            second_int = decoded[8:]
                            pos = struct.pack('<Q', struct.unpack('>Q', second_int)[0]) + \
                                      struct.pack('<Q', struct.unpack('>Q', first_int)[0])
                    except Exception:
                        pass
                self._action.append(Var(value=pos))
                last_pos_var = self._action.entity
            else:
                raise Exception(f"Unknown input type: {pos}")
        available_types.extend(self._action.types)
        if scope:
            available_types = [t for t in available_types if t.name.startswith(scope)]

        #--------------------------------------------------
        # Handle properties
        #--------------------------------------------------
        for name, val in named.items():
            prop = build.property_named(scope+name, available_types)

            if val is None:
                raise Exception(f"{prop}'s value is None, please provide a value for the property")

            if isinstance(val, int) and action_type == ActionType.Get:
                orig = val
                val = Vars(1)
                self._actions.append(build.eq(val, orig, approx=True))

            prop_var = to_var(val)

            if prop_var.isa(Builtins.ExternalInput):
                orig = prop_var
                prop_var = to_var(Vars(1))
                self._actions.append(build.eq(prop_var, orig, approx=True))

            if is_collection(prop_var.value):
                raise Exception("Can't set a property to a collection")

            if not prop_var.name:
                prop_var.name = prop.name
            if action_type.is_effect():
                self._graph._check_property(prop)
            else:
                self._graph._check_property(prop, unknown_cardinality=True)
            self._action.append(prop, prop_var)

        #--------------------------------------------------
        # Entities
        #--------------------------------------------------
        self._var = self._action.entity
        if self._var.type == Builtins.Unknown and len(self._action.types):
            self._var.type = self._action.types[0]
        if not is_add and (self._action.types or self._action.bindings):
            self._add_to_graph()

    def _to_var(self):
        if not self._graph._stack.contains(self._context):
            exception = VariableOutOfContextException(Errors.call_source(), self._var.name)
            raise exception from None
        return self._var

    def _add_to_graph(self):
        for action in self._actions:
            self._graph._action(action)

    def __call__(self, *args, **kwargs):
        pass

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            return super().__setattr__(name, value)
        exception = InvalidPropertySetException(Errors.call_source(3))
        raise exception from None

    def _make_sub(self, name: str, existing=None):
        if self._sets.get(name) is not None:
            return self._sets[name]
        if existing is not None:
            inst = InstanceProperty(self._graph, self, name, var=existing._var, scope=self._scope)
            inst._subs = existing._subs
            return inst
        prop = build.property_named(self._scope+name, self._action.types)
        if self._action.bindings.get(prop):
            return InstanceProperty(self._graph, self, name, var=self._action.bindings[prop], scope=self._scope)
        return InstanceProperty(self._graph, self, name, scope=self._scope)

    def set(self, *args, **kwargs):
        if self._graph._stack.active() is self._context:
            self._sets.update(kwargs)
        Instance(self._graph, ActionType.Bind, [self, *args], kwargs, var=self._var, scope=self._scope)
        return self

    def persist(self, *args, **kwargs):
        Instance(self._graph, ActionType.Persist, [self, *args], kwargs, var=self._var, scope=self._scope)
        return self

    def unpersist(self, *args, **kwargs):
        Instance(self._graph, ActionType.Unpersist, [self, *args], kwargs, var=self._var, scope=self._scope)
        return self

    def has_value(self):
        return self != rel.Missing # type:ignore

#--------------------------------------------------
# InstanceProperty
#--------------------------------------------------

class InstanceProperty(Producer):
    def __init__(self, graph:'Graph', instance:Instance, name:str, var=None, scope:str=""):
        super().__init__(graph)
        self._instance = instance

        self._prop = build.property_named(scope+name, instance._action.types)
        self._var = var or Var(self._prop.type, name=name)
        self._check_context()
        self._scope = scope
        new = Instance(self._graph, ActionType.Get, [instance], {name: self._var}, scope=self._scope)
        self._action = new._action

    def _check_context(self):
        if not self._graph._stack.contains(self._instance._context):
            name = f"{self._instance._var.name}.{self._var.name}"
            exception = VariableOutOfContextException(Errors.call_source(), name, is_property=True)
            raise exception from None

    def __call__(self, *args, **kwargs):
        name = f"{self._instance._var.name}.{self._var.name}"
        exception = NonCallablePropertyException(Errors.call_source(), name)
        raise exception from None

    def _make_sub(self, name: str, existing=None):
        if existing is not None and self._graph._stack.is_active(existing._instance._context):
            return existing
        return getattr(Instance(self._graph, ActionType.Get, [self], {}), name)

    def _to_var(self):
        self._check_context()
        return self._var

    def or_(self, other):
        self._graph._remove_action(self._action)
        default = build.call(Builtins.Default, [self._prop, other, self._instance, self])
        self._graph._action(default)
        return self

    def in_(self, others):
        other_rel = InlineRelation(self._graph, [(x,) for x in others])
        return self == other_rel

    def has_value(self):
        return self != rel.Missing # type:ignore

    def _remove_if_unused(self):
        # When calling append/extend we aren't necessarily doing a get on the property,
        # but we will already have added one. If we're the only thing using this get,
        # we remove it so that it doesn't unnecessarily constrain the query.
        remove = False
        for item in reversed(self._graph._stack.items):
            if item is self._action:
                remove = True
                break
            elif isinstance(item, Action):
                if self._var in item.vars():
                    remove = False
                    break
        if remove:
            self._graph._remove_action(self._action)

    def set(self, *args, **kwargs):
        return Instance(self._graph, ActionType.Get, [self], {}).set(*args, **kwargs)

    def add(self, other):
        self._remove_if_unused()
        self._graph._check_property(self._prop, multi_valued=True)
        rel = Action(ActionType.Bind, to_var(self._instance), [], {self._prop: to_var(other)})
        self._graph._action(rel)

    def extend(self, others):
        self._remove_if_unused()
        self._graph._check_property(self._prop, True)
        for other in others:
            rel = Action(ActionType.Bind, to_var(self._instance), [], {self._prop: to_var(other)})
            self._graph._action(rel)

    def choose(self, num, unique=True):
        if num < 1:
            raise ValueError("Must choose a positive number of items")
        self._remove_if_unused()
        items = [getattr(Instance(self._graph, ActionType.Get, [self._instance], {}), self._prop.name) for ix in range(num)]
        if unique:
            for ix in range(num-1):
                items[ix] < items[ix+1]
        return items

#--------------------------------------------------
# Expression
#--------------------------------------------------

class Expression(Producer):
    def __init__(self, graph:'Graph', op:mType|Task, args:List[Any]):

        super().__init__(graph)
        self._var = None
        self._context = graph._stack.active()

        # For calls to tasks with known signatures, normalize their arguments by
        # throwing on missing inputs or constructing vars for missing outputs
        if op.properties and not op.isa(Builtins.Anonymous):
            for prop, arg in zip_longest(op.properties, args):
                if arg is None:
                    if prop.is_input:
                        raise TypeError(f"{op.name} is missing a required argument: '{prop.name}'")
                    else:
                        args.append(Var(prop.type, name=prop.name))

            # Expose the last output as the result, to ensure we don't double-create it in _use_var.
            # @NOTE: Literal values like 1 show up here from calls like `rel.range(0, len(df), 1)`
            if not op.properties[-1].is_input and isinstance(args[-1], Var):
                self._var = args[-1]

        self._expr = build.call(op, args)
        self._graph._action(self._expr)

    def __call__(self, *args, **kwargs):
        raise Exception("Expressions can't be called")

    def _use_var(self):
        if self._expr.entity.isa(Builtins.Filter):
            if isinstance(runtime_env, SnowbookEnvironment):
                return
            raise FilterAsValue(Errors.call_source())
        elif not self._var:
            self._var = Var(Builtins.Unknown)
            prop = build.property_named("result", self._expr.types)
            self._expr.append(prop, self._var)

        if not self._graph._stack.contains(self._context):
            exception = VariableOutOfContextException(Errors.call_source(), self._var.name or "a result")
            raise exception from None

    def _make_sub(self, name: str, existing=None):
        if existing is not None and existing._instance._context is self._graph._stack.active():
            return existing
        return getattr(Instance(self._graph, ActionType.Get, [self], {}), name)

    def has_value(self):
        return self != rel.Missing # type:ignore

#--------------------------------------------------
# RelationNS
#--------------------------------------------------

unsafe_symbol_pattern = re.compile(r"[^a-zA-Z0-9_]", re.UNICODE)
def safe_symbol(name: str):
    return f':"{name}"' if unsafe_symbol_pattern.search(name) else f":{name}"

class RelationNS():
    def __init__(self, ns:List[str], name:str, use_rel_namespaces=False, tags:List[mType]=[]):
        if name == "getdoc":
            rich.print("[red bold]GETDOC CALLED")
            traceback.print_stack()
            return
        self._name = name
        self._ns = ns
        self._subs = {}
        self._use_rel_namespaces = use_rel_namespaces
        self._rel = self._build_rel()
        self._tags = tags

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        op = self._rel
        for t in self._tags:
            tag(self, t)
        self._ensure_args(len(args))
        return Expression(get_graph(), op, list(args))

    def __getattr__(self, name: str) -> RelationNS:
        self._subs[name] = self._make_sub(name, self._subs.get(name))
        return self._subs[name]

    def _make_sub(self, name: str, existing=None):
        if existing is not None:
            return existing
        ns = self._ns[:]
        if self._name:
            ns.append(self._name)
        return self.__class__(ns, name, use_rel_namespaces=self._use_rel_namespaces, tags=self._tags)

    def _build_rel(self, arg_count = 0):
        fqn_parts = self._ns + [self._name]
        if self._use_rel_namespaces:
            return build.relation('::'+'::'.join(fqn_parts), arg_count)
        if len(fqn_parts) == 1:
            return build.relation(fqn_parts[0], arg_count)
        else:
            return build.relation(f"{fqn_parts[0]}[{', '.join(safe_symbol(part) for part in fqn_parts[1:])}]", arg_count)

    def _ensure_args(self, arg_count):
        if len(self._rel.properties) <= arg_count:
            self._rel.properties = [Builtins.Relation.properties[i] for i in range(arg_count)]

    def add(self, *args):
        op = self._rel
        self._ensure_args(len(args))
        get_graph()._action(build.relation_action(ActionType.Bind, op, list(args)))

    def _to_var(self):
        return Var(Builtins.Relation, value=self._build_rel())

    def _tagged(self, *tags: mType):
        return self.__class__(self._ns, self._name, use_rel_namespaces=self._use_rel_namespaces, tags=list(tags))

#--------------------------------------------------
# RawRelation
#--------------------------------------------------

class RawRelation(Producer):
    def __init__(self, graph:'Graph', name:str, arity:int):
        super().__init__(graph)
        self._name = name
        self._arity = arity
        self._type = build.relation(self._name, self._arity)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return Expression(self._graph, self._type, list(args))

    def add(self, *args):
        self._graph._action(build.relation_action(ActionType.Bind, self._type, list(args)))

    def _make_sub(self, name: str, existing=None):
        return existing

    def _to_var(self):
        return Var(Builtins.Relation, value=self._type)

#--------------------------------------------------
# InlineRelation
#--------------------------------------------------

class InlineRelation():
    def __init__(self, graph:'Graph', data:List[Tuple]):
        self._var = Var(type=Builtins.InlineRawData)
        self._graph = graph
        cols = [[] for _ in range(len(data[0]))]
        for row in data:
            for i, val in enumerate(row):
                cols[i].append(to_var(val))

        params = [Var(value=col) for col in cols]
        params.append(self._var)
        q = build.relation_action(ActionType.Get, Builtins.InlineRawData, params)
        self._graph._action(q)

    def _to_var(self):
        return self._var

#--------------------------------------------------
# Rows
#--------------------------------------------------

class Rows():
    def __init__(self, graph:'Graph', data:List[Any]):
        self._graph = graph

        col_names = []
        data_cols = []

        size = len(data)
        warning_size = 1000
        error_size = 10000
        if size > error_size:
            raise errors.RowLiteralTooLarge(size, max_size=error_size)
        elif size > warning_size:
            errors.RowLiteralTooLargeWarning(size, max_size=warning_size)

        first = data[0]
        if isinstance(first, dict):
            col_names = list(first.keys())
            data_cols = [[] for _ in range(len(col_names))]
            for row in data:
                for ix, col in enumerate(col_names):
                    if col not in row:
                        raise errors.RowLiteralMismatch("keys")
                    data_cols[ix].append(row[col])
        elif isinstance(first, tuple):
            data_cols = [[] for _ in range(len(first))]
            for row in data:
                if len(row) != len(first):
                    raise errors.RowLiteralMismatch("length")
                for ix, col in enumerate(row):
                    data_cols[ix].append(col)
        elif isinstance(first, list):
            data_cols = [data]
        else:
            raise Exception("Unsupported data type: must be a list")

        data_vars = [Var(value=col) for col in data_cols]
        if col_names:
            vars = [Var(name=name) for name in col_names]
        else:
            vars = [Var() for _ in range(len(data_vars))]

        self._fetch = build.relation_action(ActionType.Get, Builtins.RawData, [*data_vars, *vars])
        self._var = vars[-1]
        self._vars = vars
        self._columns = col_names

    def _add_to_graph(self):
        if not self._graph._in_rule():
            raise errors.OutOfContextException()
        self._graph._action(self._fetch, dedupe=True)

    def _to_var(self):
        self._add_to_graph()
        return self._var

    def __getitem__(self, index):
        self._add_to_graph()
        if isinstance(index, str):
            return getattr(self, index)
        return self._vars[index]

    def __getattr__(self, name):
        if name in self._columns:
            self._add_to_graph()
            return self._vars[self._columns.index(name)]
        raise AttributeError(f"Set has no attribute '{name}'")

#--------------------------------------------------
# Symbol
#--------------------------------------------------

class Symbol():
    def __init__(self, name:str):
        self._var = Var(Builtins.Symbol, value=name)

    def _to_var(self):
        return self._var

#--------------------------------------------------
# RelationRef
#--------------------------------------------------

class RelationRef(Producer):
    def __init__(self, graph:'Graph', rel:Task|mType, args:List[Var]):
        super().__init__(graph)
        self._rel = rel
        self._args = args
        self._var = args[-1]
        self._action = build.relation_action(ActionType.Get, self._rel, self._args)

    def _use_var(self):
        self._graph._action(self._action)

    def _make_sub(self, name: str, existing=None):
        return getattr(Instance(self._graph, ActionType.Get, [self], {}), name)

    def __enter__(self):
        super().__enter__()
        self._use_var()
        Instance(self._graph, ActionType.Get, [self], {}).has_value()

#--------------------------------------------------
# Export
#--------------------------------------------------

allowed_export_types = [Type, str, numbers.Number, datetime.datetime, datetime.date, bool]

def check_type(name, type):
    if not any(isinstance(type, t) or (inspect.isclass(type) and issubclass(type, t))
                for t in allowed_export_types):
        raise TypeError(f"Argument '{name}' is an unsupported type: {type}")

def export(model, schema, kwargs):
    try:
        parser = IdentityParser(schema)
        if parser.entity:
            raise Exception(f"Schema '{schema}' is invalid. Did you mean '{parser.db}.{parser.schema}'?")
    except ParseError as e:
        raise Exception(f"Not able to parse provided schema '{schema}'. {e}")

    def decorator(func):
        # Get type hints of the function
        hints = get_type_hints(func)
        input_types = [hints[name] for name in hints if name != 'return']
        arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
        for name in arg_names:
            if name not in hints:
                raise TypeError(f"Argument '{name}' must have a type hint")
            check_type(name, hints[name])

        output_types = []
        has_return_hint = 'return' in hints
        if has_return_hint:
            ret = hints.get('return')
            if typing.get_origin(ret) is tuple:
                for t in typing.get_args(ret):
                    check_type("return", t)
                    output_types.append(t)
            else:
                check_type("return", ret)
                output_types.append(ret)
        name = f"{schema}.{func.__name__}" if schema else func.__name__
        props = {
            "name": name,
            "outputs": output_types if has_return_hint else None,
        }
        ctx = Context(model, exec_type=TaskExecType.Procedure, props=props, format="snowpark", **kwargs)
        with ctx as ret:
            inputs = to_list(Vars(len(arg_names)))
            for i in inputs:
                i._var.type = Builtins.ExternalInput
            props["inputs"] = list(zip(arg_names, [to_var(i) for i in inputs], input_types))
            outs = to_list(func(*inputs))
            if has_return_hint and len(outs) != len(output_types):
                raise TypeError(f"Function '{func.__name__}' must return {len(output_types)} values according to the type hints")
            ret(*outs)

        def wrapper():
            raise Exception("Exports can't be called directly. They are exported to the underlying platform")

        return wrapper
    return decorator

#--------------------------------------------------
# RuleStack
#--------------------------------------------------

class RuleStack():
    def __init__(self, graph:'Graph'):
        self.items = []
        self.stack = []
        self._graph = graph

    def push(self, item):
        self.stack.append(item)
        self.items.append(("push", item))

    def pop(self, item, globalize=False):
        self.stack.pop()
        self.items.append(("pop", item))
        if len(self.stack) == 0:
            compacted = self.compact()
            self.items.clear()
            if len(compacted.items):
                return compacted
        elif globalize:
            ix = self.items.index(("push", item))
            temp_items = self.items[0:ix]
            self.items = self.items[ix:]
            compacted = self.compact()
            self.items = temp_items
            if len(compacted.items):
                return compacted

    def push_item(self, item):
        if not len(self.stack):
            raise Exception("Can't push a non-context item onto an empty stack")
        self.items.append(item)

    def contains(self, item):
        for i in self.stack:
            if i is item:
                return True

    def is_active(self, item):
        if isinstance(item, Context) and item._task.behavior != Behavior.Query:
            return False
        return item is self.active()

    def in_context(self, item):
        if isinstance(item, Context) and item._task.behavior != Behavior.Query:
            return False
        return self.contains(item)

    def active(self):
        try:
            cur = self.stack[-1]
            if cur is self._graph._temp_rule:
                exception = OutOfContextException(Errors.call_source())
                raise exception from None
            return cur
        except IndexError:
            exception = OutOfContextException(Errors.call_source())
            raise exception from None

    def _expression_start(self, buffer, single_use_vars):
        consume_from = -1
        if not len(buffer):
            return consume_from
        # we can only pull vars if their only use is for this condition
        used_vars = set(buffer[-1].requires_provides()[0] & single_use_vars)
        # walk buffer in reverse collecting vars in the action until we get one
        # that doesn't provide a var we care about
        for action in reversed(buffer[:-1]):
            if not isinstance(action, Action):
                break
            req, provs, _ = action.requires_provides()
            # don't pull in vars the represent root entities even though they're provided
            # by gets. This prevents scenarios where p = Person() would get pulled in if you
            # did with p.age > 10:
            provs = provs - {action.entity}
            if len(used_vars.intersection(provs)):
                used_vars.update(req & single_use_vars)
                consume_from -= 1
            else:
                break
        return consume_from

    def compact(self) -> Task:
        stack:List[Task] = []
        buffer = []

        var_uses = {}
        for item in self.items:
            if isinstance(item, Action):
                if item.action == ActionType.Get:
                    for var in item.vars():
                        var_uses[var] = var_uses.get(var, 0) + 1
                else:
                    for var in item.vars():
                        var_uses[var] = var_uses.get(var, 0) - 1

        # check for 2 refs - one create and one use
        single_use_vars = set([var for var, uses in var_uses.items() if uses >= 0])

        for item in self.items:
            if not isinstance(item, tuple):
                buffer.append(item)
                continue

            op, value = item
            if op == "push":
                if isinstance(value, Context):
                    if len(buffer):
                        stack[-1].items.extend(buffer)
                        buffer.clear()
                    task = value._task
                elif isinstance(value, RelationRef):
                    if len(buffer):
                        stack[-1].items.extend(buffer)
                        buffer.clear()
                    task = Task()

                elif isinstance(value, Producer):
                    consume_from = self._expression_start(buffer, single_use_vars)
                    stack[-1].items.extend(buffer[:consume_from])
                    buffer = buffer[consume_from:]
                    task = Task()
                else:
                    raise Exception(f"Unknown push type: {type(value)}")

                stack.append(task)

            elif op == "pop":
                cur = stack.pop()
                cur.items.extend(buffer)
                buffer.clear()
                if not len(stack):
                    return cur
                if isinstance(value, Context) and value._op:
                    stack[-1].items.append(build.call(value._op, [Var(value=value._args), Var(Builtins.Task, value=cur)]))
                else:
                    stack[-1].items.append(build.call(cur, list(cur.bindings.values())))

        raise Exception("No task found")

#--------------------------------------------------
# Graph
#--------------------------------------------------

locals = threading.local()
locals.graph_stack = []

def get_graph() -> 'Graph':
    _ensure_stack()
    if not len(locals.graph_stack):
        raise Exception("Outside of a model context")
    return locals.graph_stack[-1]

def _ensure_stack():
    if not hasattr(locals, "graph_stack"):
        locals.graph_stack = []
    return locals.graph_stack

rel = RelationNS([], "")
global_ns = RelationNS([], "", use_rel_namespaces=True)

def alias(ref:Any, name:str):
    var = to_var(ref)
    var.name = name
    return var

def tag(ref:Any, tag:mType):
    if isinstance(ref, RelationNS):
        parents = ref._rel.parents
    elif isinstance(ref, Type):
        parents = ref._type.parents
    elif isinstance(ref, Property):
        parents = ref._type.parents
    else:
        return
    if tag not in parents:
        parents.append(tag)

def create_var() -> Instance:
    return Instance(get_graph(), ActionType.Get, [], {}, Var(Builtins.Unknown))

def create_vars(count: int) -> List[Instance]:
    return [create_var() for _ in range(count)]

def Vars(count: int):
    if count == 1:
        return create_var()
    return create_vars(count)

class Graph:
    def __init__(self, client:Client, name:str, format:str="default"):
        self.name = name
        self._executed = []
        self._client = client
        self._config = client.resources.config
        self.resources = client.resources
        self._prop_is_multi:Dict[str, bool] = {}
        self._format = format
        self._stack = RuleStack(self)
        self._restore_temp()

        self._Error = Type(self, "Error", scope="pyrel_error_")
        self._error_props = set()


    #--------------------------------------------------
    # Rule stack
    #--------------------------------------------------

    def _flush_temp(self):
        if self._temp_rule:
            self._pop(self._temp_rule, is_temp=True)
            if not len(_ensure_stack()):
                _ensure_stack().append(self)
            self._temp_rule = None

    def _restore_temp(self):
        self._temp_rule = Context(self)
        _ensure_stack().append(self)
        self._stack.push(self._temp_rule)

    def _temp_is_active(self):
        return self._temp_rule and len(self._stack.items) > 1

    def _in_rule(self):
        return not self._temp_rule

    def _push(self, item):
        _ensure_stack().append(self)
        self._flush_temp()
        self._stack.push(item)

    def _pop(self, item, exec=True, is_temp=False, globalize=False):
        _ensure_stack().pop()
        task = self._stack.pop(item, globalize=globalize)
        try:
            if exec and task:
                self._exec(item, task)
        finally:
            if not is_temp and not len(self._stack.stack):
                self._restore_temp()

    def _action(self, action:Action|List[Action], dedupe=False):
        if isinstance(action, list):
            for a in action:
                self._action(a, dedupe=dedupe)
            return
        if dedupe and self._stack.contains(action):
            return
        self._stack.push_item(action)

    def _remove_action(self, action):
        self._stack.items.remove(action)

    def _exec(self, context:Context, task):
        props = context._props
        if context._exec_type == TaskExecType.Rule:
            self._client.install(f"rule{len(self._executed)}", context._task)
        elif context._exec_type == TaskExecType.Query:
            context.results = self._client.query(context._task, tag=context._tag, format=context._format)
        elif context._exec_type == TaskExecType.Procedure:
            self._client.export_udf(props["name"], props["inputs"], props["outputs"], context._task, props.get("engine"))
        elif context._exec_type == TaskExecType.Export:
            self._client.export_table(props["name"], props["rel"], props["columns"], context._task, engine=props.get("engine"), refresh=props.get("refresh"))
        self._executed.append(context)

    #--------------------------------------------------
    # Property handling
    #--------------------------------------------------

    def _check_property(self, prop:mProperty, multi_valued=False, unknown_cardinality=False):
        name = prop.name
        if name in RESERVED_PROPS:
            exception = ReservedPropertyException(Errors.call_source(), name)
            raise exception from None

        if unknown_cardinality:
            if name in self._prop_is_multi and self._prop_is_multi[name] and Builtins.MultiValued not in prop.parents:
                prop.parents.append(Builtins.MultiValued)
            return

        if name in self._prop_is_multi:
            if self._prop_is_multi[name] != multi_valued:
                raise Exception(
                    f"Trying to use a property `{name}` as both singular and multi-valued. "
                    "This often happens when multiple types have a property with the same name, but one is single-valued and the other is multi-valued. "
                    "Consider using a plural name for the multi-valued property. "
                    "See https://relational.ai/docs/develop/core-concepts#setting-object-properties for more information."
                )
            elif self._prop_is_multi[name] and Builtins.MultiValued not in prop.parents:
                prop.parents.append(Builtins.MultiValued)
        else:
            self._prop_is_multi[name] = multi_valued

        if not multi_valued and Builtins.FunctionAnnotation not in prop.parents:
            prop.parents.append(Builtins.FunctionAnnotation)
        elif multi_valued and Builtins.MultiValued not in prop.parents:
            prop.parents.append(Builtins.MultiValued)

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------

    def Type(self, name:str, source=None):
        if source:
            return self.resources.to_model_type(self, name, source)
        return Type(self, name)

    def _error_like(self, message:str, kwargs:Dict[str, Any], is_error:bool):
        kwargs["message"] = message
        kwargs["severity"] = "error" if is_error else "warning"
        source = Errors.call_source()
        id = 0
        if source:
            id = len(errors.ModelError.error_locations)
            errors.ModelError.error_locations[id] = source
            kwargs["pyrel_id"] = id
        for k, v in kwargs.items():
            if k not in self._error_props:
                self._error_props.add(k)
                with self.rule():
                    e = self._Error()
                    rel.output.pyrel_error.add(e, k, getattr(e, k))
        if is_error:
            with self.case():
                self._Error(pyrel_id=id)
                rel.abort.add()
        return self._Error.add(**kwargs)

    def error(self, message:str, **kwargs):
        return self._error_like(message, kwargs, is_error=True)

    def warn(self, message:str, **kwargs):
        return self._error_like(message, kwargs, is_error=False)

    def rule(self, **kwargs):
        return Context(self, **kwargs)

    def scope(self, **kwargs):
        return Context(self, **kwargs)

    def case(self, **kwargs):
        return Context(self, **kwargs)

    def match(self, multiple=False, **kwargs):
        if not multiple:
            return Context(self, behavior=Behavior.OrderedChoice, **kwargs)
        else:
            return Context(self, behavior=Behavior.Union, **kwargs)

    def query(self, **kwargs):
        return Context(self, exec_type=TaskExecType.Query, **kwargs)

    def export(self, object:str = "", **kwargs):
        return export(self, object, kwargs)

    def found(self, **kwargs):
        return Context(self, op=Builtins.Exists, **kwargs)

    def not_found(self, **kwargs):
        return Context(self, op=Builtins.Not, **kwargs)

    def union(self, **kwargs):
        return Context(self, behavior=Behavior.Union, **kwargs)

    def ordered_choice(self, **kwargs):
        return Context(self, behavior=Behavior.OrderedChoice, **kwargs)

    def read(self, name:str, **kwargs):
        from relationalai.loaders.loader import read_resource_context # We do the late import to break an dependency cycle
        return read_resource_context(self, name, **kwargs)

    def load_raw(self, path:str):
        if os.path.isfile(path):
            if path.endswith('.rel'):
                self._client.load_raw_file(path)
        elif os.path.isdir(path):
            for root, _, files in os.walk(path):
                for file in files:
                    if file.endswith('.rel'):
                        file_path = os.path.join(root, file)
                        self._client.load_raw_file(file_path)

    def exec_raw(self, code:str, readonly=True, raw_results=True, abort_on_error=True, inputs={}):
        try:
            return self._client.exec_raw(code, readonly=readonly, raw_results=raw_results, inputs=inputs, abort_on_error=abort_on_error)
        except KeyboardInterrupt as e:
            print("Canceling transactions...")
            self.resources.cancel_pending_transactions()
            raise e
        except RAIException as e:
            raise e.clone() from None

    def install_raw(self, code:str, name:str|None = None, overwrite=False):
        self._client.install_raw(code, name, overwrite)




from ..metamodel import ActionType, Behavior, Task, Action
from ..compiler import Pass

#--------------------------------------------------
# WhyMeta
#--------------------------------------------------

class WhyMeta():

    def __init__(self, task:Task):
        self.task = task

#--------------------------------------------------
# WhyNot
#--------------------------------------------------

class WhyNot(Pass):

    def __init__(self, task):
        super().__init__(copying=False)
        self.why_query = Task(behavior=Behavior.Sequence)
        self.why_meta = WhyMeta(task)
        self.walk(task)

    def query(self, query: Task, parent=None):
        for item in query.items:
            if item.action == ActionType.Get:
                self.explode_get(item, parent)

    def explode_get(self, action: Action, parent=None):
        # create a query for each subset of the parts on the get
        pass


from __future__ import annotations
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
import re
from typing import Any, Sequence as PySequence
import itertools
from more_itertools import peekable

from relationalai import debugging
from relationalai.early_access.metamodel import factory as f, ir, builtins, types
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.rel.executor import RelExecutor
from relationalai.environments import runtime_env

#--------------------------------------------------
# Globals
#--------------------------------------------------

_global_id = peekable(itertools.count(0))
FORCE_DRY_RUN = False

#--------------------------------------------------
# Root tracking
#--------------------------------------------------

_track_default = False
_track_roots = ContextVar('track_roots', default=_track_default)
_global_roots = ordered_set()

def _add_root(root):
    if _track_roots.get() or isinstance(root, Fragment):
        _global_roots.add(root)

def _remove_roots(items: PySequence[Producer|Fragment]):
    for item in items:
        if hasattr(item, "__hash__") and item.__hash__ and item in _global_roots:
            _global_roots.remove(item)

# decorator
def roots(enabled=_track_default):
    def decorator(func):
        def wrapper(*args, **kwargs):
            token = _track_roots.set(enabled)
            try:
                return func(*args, **kwargs)
            finally:
                _track_roots.reset(token)
        return wrapper
    return decorator

# with root_tracking(enabled=False): ...
@contextmanager
def root_tracking(enabled=_track_default):
    token = _track_roots.set(enabled)
    try:
        yield
    finally:
        _track_roots.reset(token)

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def flatten(items:PySequence[Any]) -> list[Any]:
    flat = []
    for item in items:
        if isinstance(item, (list, tuple)):
            flat.extend(flatten(item))
        else:
            flat.append(item)
    return flat

def find_subjects(items: PySequence[Producer]) -> set[Concept|Ref]:
    subjects = set()
    for item in items:
        if isinstance(item, Concept):
            subjects.add(item)
        elif isinstance(item, ConceptExpression):
            subjects.add(item._op)
        elif isinstance(item, Expression):
            subjects.update(find_subjects(item._params))
        elif isinstance(item, Ref):
            subjects.add(item)
        elif isinstance(item, Relationship) and item._parent:
            subjects.update(find_subjects([item._parent]))
    return subjects

def to_type(item: Any) -> Concept|None:
    if isinstance(item, Concept):
        return item
    elif isinstance(item, (Ref, Alias)):
        return to_type(item._thing)
    elif isinstance(item, ConceptExpression):
        return to_type(item._op)
    elif isinstance(item, Expression):
        return to_type(item._params[-1])

def to_name(item:Any) -> str:
    if isinstance(item, Relationship) and isinstance(item._parent, Concept):
        return f"{item._parent._name}_{item._name}"
    elif isinstance(item, (Ref, Alias)):
        return item._name or to_name(item._thing)
    elif isinstance(item, RelationshipRef):
        return item._relationship._name
    return getattr(item, "_name", "v")

def find_model(items: Any) -> Model|None:
    if isinstance(items, (list, tuple)):
        for item in items:
            if hasattr(item, "_model") and item._model:
                return item._model
    else:
        if hasattr(items, "_model") and items._model:
            return items._model
    return None

def find_keys(item: Any, keys:OrderedSet[Any]|None = None) -> OrderedSet[Any]:
    if keys is None:
        keys = ordered_set()
    if isinstance(item, (list, tuple)):
        for it in item:
            find_keys(it, keys)
    elif isinstance(item, Relationship) and item._parent:
        find_keys(item._parent, keys)
    elif isinstance(item, Concept):
        if not item._is_primitive():
            keys.add(item)
    elif isinstance(item, ConceptExpression):
        pass
    elif isinstance(item, Ref):
        if isinstance(item._thing, Concept):
            if not item._thing._is_primitive():
                keys.add(item)
        else:
            find_keys(item._thing, keys)
    elif isinstance(item, RelationshipRef):
        find_keys(item._parent, keys)
    elif isinstance(item, Alias):
        find_keys(item._thing, keys)
    elif isinstance(item, Aggregate):
        find_keys(item._args, keys)
    elif isinstance(item, Expression):
        find_keys(item._params, keys)
    elif isinstance(item, BranchRef):
        find_keys(item._match, keys)
    elif isinstance(item, Match):
        pass
    elif isinstance(item, Distinct):
        pass
    elif isinstance(item, (str, int, float, bool)):
        pass
    else:
        raise ValueError(f"Cannot find keys for {item}")
    return keys

def with_source(item:Any):
    if not hasattr(item, "_source"):
        raise ValueError(f"Item {item} has no source")
    if debugging.DEBUG:
        source = item._source.to_source_info()
        if source:
            return { "file": source.file, "line": source.line, "source": source.source }
        else:
            return {"file":item._source.file, "line":item._source.line}
    else:
        return {"file":item._source.file, "line":item._source.line}


#--------------------------------------------------
# Producer
#--------------------------------------------------

class Producer:
    def __init__(self, model:Model|None) -> None:
        self._id = next(_global_id)
        self._model = model

    #--------------------------------------------------
    # Infix operator overloads
    #--------------------------------------------------

    def _bin_op(self, op, left, right) -> Expression:
        res = Concept.builtins["Number"].ref("res")
        return Expression(Relationship.builtins[op], left, right, res)

    def __add__(self, other):
        return self._bin_op("+", self, other)
    def __radd__(self, other):
        return self._bin_op("+", other, self)

    def __mul__(self, other):
        return self._bin_op("*", self, other)
    def __rmul__(self, other):
        return self._bin_op("*", other, self)

    def __sub__(self, other):
        return self._bin_op("-", self, other)
    def __rsub__(self, other):
        return self._bin_op("-", other, self)

    def __truediv__(self, other):
        return self._bin_op("/", self, other)
    def __rtruediv__(self, other):
        return self._bin_op("/", other, self)

    def __floordiv__(self, other):
        return self._bin_op("//", self, other)
    def __rfloordiv__(self, other):
        return self._bin_op("//", other, self)

    def __pow__(self, other):
        return self._bin_op("^", self, other)
    def __rpow__(self, other):
        return self._bin_op("^", other, self)

    def __mod__(self, other):
        return self._bin_op("%", self, other)
    def __rmod__(self, other):
        return self._bin_op("%", other, self)

    def __neg__(self):
        return self._bin_op("*", self, -1)

    #--------------------------------------------------
    # Filter overloads
    #--------------------------------------------------

    def _filter(self, op, left, right) -> Expression:
        return Expression(Relationship.builtins[op], left, right)

    def __gt__(self, other):
        return self._filter(">", self, other)
    def __ge__(self, other):
        return self._filter(">=", self, other)
    def __lt__(self, other):
        return self._filter("<", self, other)
    def __le__(self, other):
        return self._filter("<=", self, other)
    def __eq__(self, other) -> Any:
        return self._filter("=", self, other)
    def __ne__(self, other) -> Any:
        return self._filter("!=", self, other)

    #--------------------------------------------------
    # And/Or
    #--------------------------------------------------

    def __or__(self, other) -> Match:
        return Match(self, other)

    def __and__(self, other) -> Union:
        return Union(self, other)

    #--------------------------------------------------
    # Relationship handling
    #--------------------------------------------------

    def _get_relationship(self, name:str) -> Relationship|RelationshipRef:
        root_type = to_type(self) or Concept.builtins["Any"]
        r = Relationship(
            f"{{{root_type}}} has {{{name}:Any}}",
            parent=self, short_name=name, model=self._model
        )
        return r

    #--------------------------------------------------
    # getattr
    #--------------------------------------------------

    def __getattr__(self, name:str) -> Any:
        if name.startswith("_"):
            raise AttributeError(f"{type(self).__name__} has no attribute {name}")
        if not hasattr(self, "_relationships"):
            return super().__getattribute__(name)
        if name not in self._relationships:
            self._relationships[name] = self._get_relationship(name)
        return self._relationships[name]

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif isinstance(value, Relationship):
            value._parent = self
            value._short_name = name
            self._relationships[name] = value
        else:
            raise AttributeError(f"Cannot set attribute {name} on {type(self).__name__}")

    #--------------------------------------------------
    # ref + alias
    #--------------------------------------------------

    def ref(self, name:str|None=None) -> Ref:
        return Ref(self, name=name)

    def alias(self, name:str) -> Alias:
        return Alias(self, name)

    #--------------------------------------------------
    # Find model
    #--------------------------------------------------

    def _find_model(self, items:list[Any]) -> Model|None:
        if self._model:
            return self._model

        for item in items:
            if isinstance(item, (Producer, Fragment)) and item._model:
                self._model = item._model
                return item._model
        return None

    #--------------------------------------------------
    # Hash
    #--------------------------------------------------

    __hash__ = object.__hash__

    #--------------------------------------------------
    # _pprint
    #--------------------------------------------------

    def _pprint(self, indent:int=0) -> str:
        return str(self)

    #--------------------------------------------------
    # Fallbacks
    #--------------------------------------------------

    def select(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.select` not implemented")

    def where(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.where` not implemented")

    def require(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.require` not implemented")

    def then(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.then` not implemented")

#--------------------------------------------------
# Ref
#--------------------------------------------------

class Ref(Producer):
    def __init__(self, thing:Producer, name:str|None=None):
        super().__init__(thing._model)
        self._thing = thing
        self._name = name
        self._relationships = {}

    def _get_relationship(self, name: str) -> Relationship | RelationshipRef:
        rel = getattr(self._thing, name)
        return RelationshipRef(self, rel)

    def __str__(self) -> str:
        if self._name:
            return f"{self._name}{self._id}"
        return f"{self._thing}{self._id}"

class RelationshipRef(Producer):
    def __init__(self, parent:Any, relationship:Relationship|RelationshipRef):
        super().__init__(find_model([parent, relationship]))
        self._parent = parent
        if isinstance(relationship, RelationshipRef):
            relationship = relationship._relationship
        self._relationship:Relationship = relationship
        self._relationships = {}

    def _get_relationship(self, name: str) -> Relationship|RelationshipRef:
        rel = super()._get_relationship(name)
        assert isinstance(rel, Relationship)
        return RelationshipRef(self, rel)

    def __call__(self, *args: Any) -> Any:
        clean_args = list(args)
        if len(args) < len(self._relationship._fields):
            clean_args = [self._parent, *clean_args]
        return Expression(self._relationship, *clean_args)

    def __str__(self) -> str:
        return f"{self._parent}.{self._relationship}"

#--------------------------------------------------
# Concept
#--------------------------------------------------

class Concept(Producer):
    builtins = {}

    def __init__(self, name:str, extends:list[Any] = [], model:Model|None=None):
        super().__init__(model)
        self._name = name
        self._relationships = {}
        self._extends = []
        for e in extends:
            if isinstance(e, Concept):
                self._extends.append(e)
            elif python_types_to_concepts.get(e):
                self._extends.append(python_types_to_concepts[e])
            else:
                raise ValueError(f"Unknown concept {e} in extends")

    def when(self, *args: Any) -> Fragment:
        subjects = find_subjects(args)
        if len(subjects) == 0:
            raise ValueError("No subjects found in when clause")
        elif len(subjects) > 1:
            raise ValueError("Multiple subjects found in when clause")
        return where(*args).then(self(subjects.pop()))

    def require(self, *args: Any) -> Fragment:
        return where(self).require(*args)

    def _isa(self, other:Concept) -> bool:
        if self is other:
            return True
        for parent in self._extends:
            if parent._isa(other):
                return True
        return False

    def _is_primitive(self) -> bool:
        return self._isa(Primitive)

    def __call__(self, identity:Any=None, **kwargs: Any) -> Any:
        return ConceptExpression(self, identity, kwargs)

    def __str__(self):
        return self._name

#--------------------------------------------------
# Builtin Concepts
#--------------------------------------------------

Primitive = Concept.builtins["Primitive"] = Concept("Primitive")
Error = Concept.builtins["Error"] = Concept("Error")

# Load builtin types
for builtin in types.builtin_types:
    if isinstance(builtin, ir.ScalarType):
        Concept.builtins[builtin.name] = Concept(builtin.name, extends=[Primitive])

Integer = Concept.builtins["Int"]
Decimal = Concept.builtins["Decimal"]
String = Concept.builtins["String"]
Bool = Concept.builtins["Bool"]

python_types_to_concepts = {
    int: Concept.builtins["Int"],
    float: Concept.builtins["Float"],
    str: Concept.builtins["String"],
    bool: Concept.builtins["Bool"],
}

#--------------------------------------------------
# Relationship
#--------------------------------------------------

@dataclass
class Field():
    name:str
    type_str:str
    is_many:bool = False

class Relationship(Producer):
    builtins = {}

    def __init__(self, madlib:Any, parent:Producer|None=None, short_name:str="", model:Model|None=None, fields:list[Field]|None=None, ir_relation:ir.Relation|None=None):
        super().__init__(model or find_model(parent) or find_model(args))
        self._parent = parent
        self._madlib = madlib
        self._short_name = short_name
        self._relationships = {}
        self._fields:list[Field] = fields or self._parse_schema_format(madlib)
        self._ir_relation = ir_relation

    @property
    def _name(self):
        return self._short_name or self._madlib

    def _parse_schema_format(self, format_string:str):
        # Pattern to extract fields like {Type*} or {name:Type*}
        pattern = r'\{([a-zA-Z0-9_]+)(\*?)(?::([a-zA-Z0-9_]+)(\*?))?\}'
        matches = re.findall(pattern, format_string)

        used_names = set()
        fields = []
        for field_name, field_name_is_many, field_type, field_type_is_many in matches:
            # If no type is specified, use the field name as the type
            if not field_type:
                field_type = field_name
                field_name = field_name.lower()
                field_type_is_many = field_name_is_many

            ix = 2
            while field_name in used_names:
                field_name = f"{field_name}{ix}"
                ix += 1
            used_names.add(field_name)

            fields.append(Field(field_name, field_type, bool(field_type_is_many)))

        return fields

    def __call__(self, *args: Any) -> Any:
        clean_args = list(args)
        if len(args) < len(self._fields):
            if self._parent:
                clean_args = [self._parent, *clean_args]
        return Expression(self, *clean_args)

    def __str__(self):
        if self._parent and self._short_name:
            return f"{self._parent}.{self._short_name}"
        return self._name

for builtin in builtins.builtin_relations:
    fields = []
    for field in builtin.fields:
        field_type = re.sub(r'[\]\}]', '*', re.sub(r'[\[\{\(]', '', str(field.type))).strip()
        fields.append(f"{field.name}:{field_type}")
    args = ' '.join([f"{{{f}}}"for f in fields])
    Relationship.builtins[builtin.name] =  Relationship(f"{builtin.name} {args}", parent=None, short_name=builtin.name, ir_relation=builtin)

#--------------------------------------------------
# Expression
#--------------------------------------------------

infix_ops = ["+", "-", "*", "/", "//", "^", "%", ">", ">=", "<", "<=", "=", "!="]

class Expression(Producer):
    def __init__(self, op:Relationship|Concept, *params:Any):
        super().__init__(op._model or find_model(params))
        self._op = op
        self._params = params
        self._source = runtime_env.get_source_pos()

        _add_root(self)
        _remove_roots([p for p in params if not isinstance(p, ConceptExpression)])

    def when(self, *args: Any) -> Fragment:
        return where(*args).then(self)

    def __str__(self):
        return f"({self._op} {' '.join(map(str, self._params))})"

    def _pprint(self, indent:int=0) -> str:
        if self._op._name in infix_ops:
            a, b = self._params[0], self._params[1]
            return f"{' ' * indent}{a} {self._op} {b}"
        return f"{' ' * indent}{self._op}({' '.join(map(str, self._params))})"

    def __getattr__(self, name: str):
        raise AttributeError(f"Expression has no attribute {name}")

class ConceptExpression(Expression):
    def __init__(self, con:Concept, identity:Any, kwargs:dict[str, Any]):
        super().__init__(con, identity, kwargs)
        self._relationships = {}

    def _get_relationship(self, name: str) -> Relationship|RelationshipRef:
        parent_rel = getattr(self._op, name)
        return RelationshipRef(self, parent_rel)

    def __getattr__(self, name: str):
        return Producer.__getattr__(self, name)

#--------------------------------------------------
# Aggregate
#--------------------------------------------------

class Aggregate(Producer):
    def __init__(self, op:Relationship, *args: Any):
        super().__init__(op._model or find_model(args))
        self._op = op
        self._args = args
        _remove_roots(args)
        self._group = []
        self._where = where()

    def where(self, *args: Any) -> Aggregate:
        if not self._model:
            self._model = find_model(args)
        self._where = self._where.where(*args)
        return self

    def per(self, *args: Any) -> Aggregate:
        if not self._model:
            self._model = find_model(args)
        self._group = list(args)
        return self

    def __getattr__(self, name: str):
        raise AttributeError(f"Expression has no attribute {name}")

    def __str__(self):
        args = ', '.join(map(str, self._args))
        group = ', '.join(map(str, self._group))
        where = ""
        if group:
            group = f" (per {group})"
        if self._where._where:
            items = ', '.join(map(str, self._where._where))
            where = f" (where {items})"
        return f"({self._op} {args}{group}{where})"

class Group():
    def __init__(self, *args: Any):
        self._group = args

    def __str__(self):
        args = ', '.join(map(str, self._group))
        return f"(per {args})"

    #--------------------------------------------------
    # Agg funcs
    #--------------------------------------------------

    def count(self, *args: Any) -> Aggregate:
        return Aggregate(Relationship.builtins["count"], *args).per(*self._group)

    def sum(self, *args: Any) -> Aggregate:
        return Aggregate(Relationship.builtins["sum"], *args).per(*self._group)

    def avg(self, *args: Any) -> Aggregate:
        return Aggregate(Relationship.builtins["avg"], *args).per(*self._group)

    def min(self, *args: Any) -> Aggregate:
        return Aggregate(Relationship.builtins["min"], *args).per(*self._group)

    def max(self, *args: Any) -> Aggregate:
        return Aggregate(Relationship.builtins["max"], *args).per(*self._group)

#--------------------------------------------------
# Aggregate builtins
#--------------------------------------------------

def per(*args: Any) -> Group:
    return Group(*args)

def count(*args: Any) -> Aggregate:
    return Aggregate(Relationship.builtins["count"], *args)

def sum(*args: Any) -> Aggregate:
    return Aggregate(Relationship.builtins["sum"], *args)

def avg(*args: Any) -> Aggregate:
    return Aggregate(Relationship.builtins["avg"], *args)

def min(*args: Any) -> Aggregate:
    return Aggregate(Relationship.builtins["min"], *args)

def max(*args: Any) -> Aggregate:
    return Aggregate(Relationship.builtins["max"], *args)

#--------------------------------------------------
# Alias
#--------------------------------------------------

class Alias(Producer):
    def __init__(self, thing:Producer, name:str):
        super().__init__(thing._model)
        self._thing = thing
        self._name = name
        self._relationships = {}
        _remove_roots([thing])

    def __str__(self) -> str:
        return f"{self._thing} as {self._name}"

#--------------------------------------------------
# Match
#--------------------------------------------------

class BranchRef(Producer):
    def __init__(self, match:Match, ix:int):
        super().__init__(match._model)
        self._match = match
        self._ix = ix

    def __str__(self):
        return f"{self._match}#{self._ix}"

class Match(Producer):
    def __init__(self, *args: Any):
        super().__init__(find_model(args))
        self._args = list(self._flatten_args(args))
        if any(isinstance(arg, Fragment) and arg._is_effect() for arg in self._args):
            _add_root(self)
        _remove_roots(args)

        # check for validity
        is_select = None
        ret_count = 0
        for arg in self._args:
            if isinstance(arg, Fragment) and arg._is_effect():
                if is_select:
                    raise ValueError("Cannot mix expression and effect clauses in a match")
                is_select = False
            elif isinstance(arg, Fragment) and not arg._is_effect():
                if is_select is False:
                    raise ValueError("Cannot mix effect and expression clauses in a match")
                is_select = True
                if ret_count == 0:
                    ret_count = len(arg._select)
                elif ret_count != len(arg._select):
                    raise ValueError("All clauses must have the same number of return values")
            elif isinstance(arg, (str, int, float, bool)):
                if is_select is False:
                    raise ValueError("Cannot mix then and select clauses in a match")
                is_select = True
                if ret_count == 0:
                    ret_count = 1
                elif ret_count != 1:
                    raise ValueError("All clauses must have the same number of return values")
            elif isinstance(arg, Expression):
                if is_select is None:
                    is_select = True

        self._is_select = is_select
        self._ret_count = ret_count

    def _flatten_args(self, args):
        for arg in args:
            if isinstance(arg, Match):
                for sub_arg in arg._args:
                    yield sub_arg
            else:
                yield arg

    def __iter__(self):
        for ix in range(self._ret_count):
            yield BranchRef(self, ix)

    def __str__(self):
        return " | ".join(map(str, self._args))

#---------------------------------------------/-----
# Union
#--------------------------------------------------

class Union(Match):
    def __str__(self):
        return " & ".join(map(str, self._args))

def union(*args: Any) -> Union:
    return Union(*args)

#--------------------------------------------------
# Quantifiers
#--------------------------------------------------

class Quantifier():
    def __init__(self, *args: Any):
        self._args = args
        self._model = find_model(args)
        _remove_roots(args)
        self._where = where()

    def where(self, *args: Any) -> Quantifier:
        if not self._model:
            self._model = find_model(args)
        self._where = self._where.where(*args)
        return self

class Exists(Quantifier):
    def __str__(self):
        args_str = ','.join(map(str, self._args))
        where_items = '\n        '.join(map(str, self._where._where))
        where = f"\n        {where_items}" if self._where._where else ""
        return f"(exists [{args_str}]{where})"

class Forall(Quantifier):
    def __str__(self):
        args_str = ','.join(map(str, self._args))
        where_items = '\n        '.join(map(str, self._where._where))
        where = f"\n        {where_items}" if self._where._where else ""
        return f"(forall [{args_str}]{where})"

class Not(Quantifier):
    def where(self, *args: Any) -> Quantifier:
        raise NotImplementedError("Not cannot have a where clause")

    def __str__(self):
        args_str = '\n    '.join(map(str, self._args))
        return f"(not {args_str})"

def exists(*args: Any) -> Exists:
    return Exists(*args)

def forall(*args: Any) -> Forall:
    return Forall(*args)

def not_(*args: Any) -> Not:
    return Not(*args)

#--------------------------------------------------
# Distinct
#--------------------------------------------------

class Distinct():
    def __init__(self, *args: Any):
        self._args = args
        self._model = find_model(args)
        _remove_roots(args)

def distinct(*args: Any) -> Distinct:
    return Distinct(*args)

#--------------------------------------------------
# Fragment
#--------------------------------------------------

class Fragment():
    def __init__(self, parent:Fragment|None=None, model:Model|None=None):
        self._id = next(_global_id)
        self._select = []
        self._where = []
        self._require = []
        self._then = []
        self._model = parent._model if parent else model
        self._parent = parent
        self._source = runtime_env.get_source_pos()
        self._meta = {}
        if parent:
            self._select.extend(parent._select)
            self._where.extend(parent._where)
            self._require.extend(parent._require)
            self._then.extend(parent._then)

    def _add_items(self, items:PySequence[Any], to_attr:list[Any]):
        # TODO: ensure that you are _either_ a select, require, or then
        # not a mix of them
        if to_attr is self._then or to_attr is self._require:
            _add_root(self)

        _remove_roots(items)
        to_attr.extend(items)
        if not self._model:
            self._model = find_model(items)
        return self

    def where(self, *args: Any) -> Fragment:
        f = Fragment(parent=self)
        return f._add_items(args, f._where)

    def select(self, *args: Any) -> Fragment:
        f = Fragment(parent=self)
        return f._add_items(args, f._select)

    def require(self, *args: Any) -> Fragment:
        f = Fragment(parent=self)
        return f._add_items(args, f._require)

    def then(self, *args: Any) -> Fragment:
        f = Fragment(parent=self)
        return f._add_items(args, f._then)

    def meta(self, **kwargs: Any) -> Fragment:
        self._meta.update(kwargs)
        return self

    #--------------------------------------------------
    # helpers
    #--------------------------------------------------

    def _is_effect(self) -> bool:
        return bool(self._then or self._require or (self._parent and self._parent._is_effect()))

    #--------------------------------------------------
    # And/Or
    #--------------------------------------------------

    def __or__(self, other) -> Match:
        return Match(self, other)

    def __and__(self, other) -> Union:
        return Union(self, other)

    #--------------------------------------------------
    # Stringify
    #--------------------------------------------------

    def __str__(self):
        sections = []
        if self._select:
            select = '\n    '.join(map(str, self._select))
            sections.append(f"(select\n    {select})")
        if self._where:
            where = '\n    '.join(map(str, self._where))
            sections.append(f"(where\n    {where})")
        if self._require:
            require = '\n    '.join(map(str, self._require))
            sections.append(f"(require\n    {require})")
        if self._then:
            effects = '\n    '.join(map(str, self._then))
            sections.append(f"(then\n    {effects})")

        return "\n".join(sections)

    #--------------------------------------------------
    # Execute
    #--------------------------------------------------

    def __iter__(self):
        # Iterate over the rows of the fragment's results
        return self.to_df().itertuples(index=False).__iter__()

    def inspect(self):
        # @TODO what format? maybe ignore row indices?
        print(self.to_df())

    def to_df(self):
        """Convert the fragment's results to a pandas DataFrame."""
        # @TODO currently this code assumes a Rel executor; should dispatch based on config
        if not self._model:
            raise ValueError("Cannot execute fragment without a model context")

        qb_model = self._model or Model("anon")
        ir_model = qb_model._to_ir()
        self._source = runtime_env.get_source_pos()
        with debugging.span("query", dsl=str(self), **with_source(self), meta=self._meta) as query_span:
            query_task = qb_model._compiler.fragment(self)
            results = qb_model._to_executor().execute(ir_model, query_task)
            query_span["results"] = results
            return results

#--------------------------------------------------
# Select / Where
#--------------------------------------------------

def select(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return Fragment().select(*args)

def where(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return Fragment().where(*args)

def require(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return Fragment().require(*args)

def then(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return Fragment().then(*args)

#--------------------------------------------------
# Model
#--------------------------------------------------

class Model():
    def __init__(self, name:str, dry_run: bool = False, implicit_effects:bool = False):
        self._id = next(_global_id)
        self.name = name
        self._dry_run = dry_run or FORCE_DRY_RUN
        self.concepts:dict[str, Concept] = {}
        self.relationships:list[Relationship] = []
        if implicit_effects:
            _track_roots.set(True)

        # Compiler
        self._compiler = Compiler()
        self._root_version = _global_roots.version()
        self._last_compilation = None

        # Executor
        self._executor = None

    def _to_ir(self):
        if not _global_roots.has_changed(self._root_version) and self._last_compilation:
            return self._last_compilation
        self._last_compilation = self._compiler.model(self)
        self._root_version = _global_roots.version()
        return self._last_compilation

    def _to_executor(self):
        if not self._executor:
            self._executor = RelExecutor(self.name, dry_run=self._dry_run)
        return self._executor

    def Concept(self, name:str, extends:list[Concept|Any]=[]) -> Concept:
        if name not in self.concepts:
            self.concepts[name] = Concept(name, model=self, extends=extends)
        return self.concepts[name]

    def Relationship(self, *args) -> Relationship:
        return Relationship(*args, parent=None, short_name="", model=self)

    def __enter__(self):
        self._token = _track_roots.set(True)

    def __exit__(self, exc_type, exc_value, traceback):
        _track_roots.reset(self._token)


#--------------------------------------------------
# Compile
#--------------------------------------------------

class CompilerContext():
    def __init__(self, compiler:Compiler, parent:CompilerContext|None=None):
        self.compiler = compiler
        self.parent = parent
        self.value_map:dict[Producer, ir.Value|list[ir.Var]] = parent.value_map.copy() if parent else {}
        self.items:OrderedSet[ir.Task] = OrderedSet()
        self.into_vars:list[ir.Var] = parent.into_vars.copy() if parent else []
        self.global_value_map:dict[Producer, ir.Value|list[ir.Var]] = parent.global_value_map if parent else {}

    def to_value(self, item:Producer) -> ir.Value|list[ir.Var]:
        if item not in self.value_map:
            if item in self.global_value_map:
                self.value_map[item] = self.global_value_map[item]
            else:
                name = to_name(item)
                qb_type = to_type(item)
                type = self.compiler.to_type(qb_type) if qb_type else types.Any
                name = name.lower()
                self.map_var(item, f.var(name, type))
        return self.value_map[item]

    def map_var(self, item:Producer, value:ir.Value|list[ir.Var]):
        self.global_value_map[item] = value
        self.value_map[item] = value

    def _has_item(self, item:ir.Task) -> bool:
        return bool(item in self.items or (self.parent and self.parent._has_item(item)))

    def add(self, item:ir.Task):
        if not self._has_item(item):
            self.items.add(item)

    def safe_wrap(self, required_hoists:PySequence[ir.VarOrDefault]) -> ir.Task:
        first = self.items[0]
        if len(self.items) == 1 and isinstance(first, ir.Logical) and set(required_hoists).issubset(set(first.hoisted)):
            return self.items[0]
        return f.logical(list(self.items), required_hoists)

    def clone(self):
        return CompilerContext(self.compiler, self)


class Compiler():
    def __init__(self):
        self.types:dict[Concept, ir.ScalarType] = {}
        self.name_to_type:dict[str, ir.Type] = {}
        self.relations:dict[Relationship|Concept|RelationshipRef, ir.Relation] = {}

    #--------------------------------------------------
    # Type/Relation conversion
    #--------------------------------------------------

    def to_type(self, concept:Concept) -> ir.ScalarType:
        if concept not in self.types:
            if concept._name in types.builtin_scalar_types_by_name:
                self.types[concept] = types.builtin_scalar_types_by_name[concept._name]
            else:
                parent_types = [self.to_type(parent) for parent in concept._extends]
                self.types[concept] = f.scalar_type(concept._name, parent_types)
            self.name_to_type[concept._name] = self.types[concept]
        return self.types[concept]

    def to_relation(self, item:Concept|Relationship|RelationshipRef) -> ir.Relation:
        if item not in self.relations:
            if isinstance(item, Concept):
                fields = [f.field(item._name.lower(), self.to_type(item))]
                relation = f.relation(item._name, fields)
            elif isinstance(item, Relationship):
                if item._ir_relation:
                    relation = item._ir_relation
                else:
                    fields = [f.field(cur.name, self.name_to_type.get(cur.type_str, types.Any)) for cur in item._fields]
                    relation = f.relation(item._name, fields)
            elif isinstance(item, RelationshipRef):
                relation = self.to_relation(item._relationship)
            self.relations[item] = relation
            return relation
        else:
            return self.relations[item]

    #--------------------------------------------------
    # Model
    #--------------------------------------------------

    @roots(enabled=False)
    def model(self, model:Model) -> ir.Model:
        rules = []
        with debugging.span("rule_batch"):
            for rule in _global_roots:
                if not rule._model or rule._model == model:
                    meta = rule._meta if isinstance(rule, Fragment) else {}
                    with debugging.span("rule", dsl=str(rule), **with_source(rule), meta=meta) as rule_span:
                        rule_ir = self.compile_task(rule)
                        rules.append(rule_ir)
                        rule_span["metamodel"] = str(rule_ir)
        root = f.logical(rules)
        engines = ordered_set()
        relations = OrderedSet.from_iterable(self.relations.values())
        types = OrderedSet.from_iterable(self.types.values())
        return f.model(engines, relations, types, root)

    #--------------------------------------------------
    # Compile
    #--------------------------------------------------

    @roots(enabled=False)
    def compile_task(self, thing:Expression|Fragment) -> ir.Task:
        if isinstance(thing, (Expression, Match, Union)):
            return self.root_expression(thing)
        elif isinstance(thing, Fragment):
            return self.fragment(thing)

    #--------------------------------------------------
    # Root expression
    #--------------------------------------------------

    def root_expression(self, item:Expression) -> ir.Task:
        ctx = CompilerContext(self)
        self.update(item, ctx)
        return f.logical(list(ctx.items))

    #--------------------------------------------------
    # Fragment
    #--------------------------------------------------

    def fragment(self, fragment:Fragment, parent_ctx:CompilerContext|None=None, into_vars:list[ir.Var] = []) -> ir.Task:
        ctx = CompilerContext(self, parent_ctx)
        self.where(fragment._where, ctx)
        self.require(fragment._require, ctx)
        self.then(fragment._then, ctx)
        self.select(fragment._select, ctx)
        return f.logical(list(ctx.items), ctx.into_vars)

    def where(self, items:PySequence[Expression], ctx:CompilerContext):
        for item in items:
            self.lookup(item, ctx)

    def select(self, items:PySequence[Expression], ctx:CompilerContext):
        if not items:
            return

        fields = []
        for ix, item in enumerate(items):
            keys = find_keys(item)
            for key in keys:
                self.lookup(key, ctx)
            sub_ctx = ctx.clone()
            raw_v = self.lookup(item, sub_ctx)
            if not isinstance(raw_v, list):
                raw_v = [raw_v]
            if len(sub_ctx.items) >= 1:
                ctx.add(sub_ctx.safe_wrap([ir.Default(v, None) for v in raw_v if isinstance(v, ir.Var)]))

            for v in raw_v:
                name = f"v{len(fields)}"
                if isinstance(item, Alias):
                    name = item._name
                elif isinstance(v, ir.Var):
                    name = v.name
                if ctx.into_vars:
                    ctx.add(f.lookup(builtins.eq, [ctx.into_vars[ix], v]))
                else:
                    fields.append((name, v))
        if fields:
            ctx.add(f.output(fields))

    def require(self, items:PySequence[Expression], ctx:CompilerContext):
        for item in items:
            not_ctx = ctx.clone()
            self.lookup(item, not_ctx)
            not_body = f.logical(list(not_ctx.items))

            sub_ctx = ctx.clone()
            sub_ctx.add(f.not_(not_body))
            item_str = item._pprint() if isinstance(item, Producer) else str(item)
            self.update(Error(message=f"Requirement not met: {item_str}"), sub_ctx)
            ctx.add(f.logical(list(sub_ctx.items)))

    def then(self, items:PySequence[Expression], ctx:CompilerContext):
        for item in items:
            sub_ctx = ctx.clone()
            self.update(item, sub_ctx)
            if len(sub_ctx.items) > 1:
                ctx.add(f.logical(list(sub_ctx.items)))
            elif len(sub_ctx.items) == 1:
                ctx.add(sub_ctx.items[0])

    #--------------------------------------------------
    # Lookup
    #--------------------------------------------------

    def lookup(self, item:Any, ctx:CompilerContext) -> ir.Value|list[ir.Var]:
        if isinstance(item, ConceptExpression):
            assert isinstance(item._op, Concept)
            relation = self.to_relation(item._op)
            (ident, kwargs) = item._params

            # if this has an ident, that's the only thing we need
            if ident is not None:
                return self.lookup(ident, ctx)

            # otherwise we have to construct one
            out = ctx.to_value(item)
            rels = {self.to_relation(getattr(item._op, k)): self.lookup(v, ctx)
                    for k, v in kwargs.items()}
            assert isinstance(out, ir.Var)
            ctx.add(f.construct(out, rels))
            return out

        elif isinstance(item, Expression):
            params = [self.lookup(p, ctx) for p in item._params]
            relation = self.to_relation(item._op)
            ctx.add(f.lookup(relation, flatten(params)))
            return params[-1]

        elif isinstance(item, Concept):
            v = ctx.to_value(item)
            if not item._isa(Primitive):
                assert isinstance(v, ir.Var)
                relation = self.to_relation(item)
                ctx.add(f.lookup(relation, [v]))
            return v

        elif isinstance(item, (Relationship, RelationshipRef)):
            p = self.lookup(item._parent, ctx)
            v = ctx.to_value(item)
            relation = self.to_relation(item)
            assert isinstance(p, ir.Var) and isinstance(v, ir.Var)
            ctx.add(f.lookup(relation, [p, v]))
            # TODO handle non-binary
            return v

        elif isinstance(item, Ref):
            root = item._thing
            prev_mapping = ctx.to_value(root)
            out = ctx.to_value(item)
            ctx.map_var(root, out)
            self.lookup(root, ctx)
            ctx.map_var(root, prev_mapping)
            return out

        elif isinstance(item, Alias):
            return self.lookup(item._thing, ctx)

        elif isinstance(item, Aggregate):
            relation = self.to_relation(item._op)
            assert relation.fields[0].name == "projection", "Aggregate structure has changed"

            # group
            group = [self.lookup(g, ctx) for g in item._group]
            group = [item for item in flatten(group) if isinstance(item, ir.Var)]

            # Agg context
            agg_ctx = ctx.clone()
            out = agg_ctx.to_value(item)
            assert isinstance(out, ir.Var)

            # additional wheres
            self.where(item._where._where, agg_ctx)

            # raw_args
            arg_count = len(relation.fields) - 3 # skip projection, group, and result
            raw_args = flatten([self.lookup(a, agg_ctx) for a in item._args])

            # the projection includes all keys for the args
            projection = flatten([self.lookup(key, agg_ctx) for key in find_keys(item._args)])
            # the projection is also all raw_args that aren't consumed by the agg
            projection += raw_args[:-arg_count] if arg_count else raw_args
            projection = list(dict.fromkeys([item for item in projection if isinstance(item, ir.Var)]))

            # agg args + result var
            args = raw_args[-arg_count:] if arg_count else []
            args.append(out)

            # add the aggregation
            ir_agg = f.aggregate(relation, projection, group, args)
            if agg_ctx.items:
                internal = set(raw_args) - set(ctx.value_map.values())
                ctx.add(f.logical(list(agg_ctx.items), list([ir.Default(v, None) for v in internal])))
            ctx.add(ir_agg)
            return out

        elif isinstance(item, Not):
            not_ctx = ctx.clone()
            for a in item._args:
                self.lookup(a, not_ctx)
            ctx.add(f.not_(f.logical(list(not_ctx.items))))

        elif isinstance(item, Quantifier):
            agg_ctx = ctx.clone()
            self.where(item._where._where, agg_ctx)
            args = [self.lookup(a, agg_ctx) for a in item._args]
            args = [item for item in flatten(args) if isinstance(item, ir.Var)]
            body = f.logical(list(agg_ctx.items))
            if isinstance(item, Exists):
                ctx.add(f.exists(args, body))
            elif isinstance(item, Forall):
                ctx.add(f.for_all(args, body))

        elif isinstance(item, Fragment):
            sub_ctx = ctx.clone()
            ctx.add(self.fragment(item, sub_ctx))
            if len(ctx.into_vars) == 1:
                return ctx.into_vars[0]
            elif len(ctx.into_vars) > 1:
                return ctx.into_vars
            elif len(item._select) == 1:
                return sub_ctx.to_value(item._select[0])
            elif len(item._select) > 1:
                return flatten([sub_ctx.to_value(s) for s in item._select])

        elif isinstance(item, (Match, Union)):
            branches = []
            vars = []
            if item._is_select:
                vars = ctx.value_map.get(item)
                if not vars:
                    vars = [f.var(f"v{i}") for i in range(item._ret_count)]
                    ctx.map_var(item, vars)
                assert isinstance(vars, list)
                for branch in item._args:
                    branch_ctx = ctx.clone()
                    branch_ctx.into_vars = vars
                    v = self.lookup(branch, branch_ctx)
                    if not isinstance(v, list):
                        v = [v]
                    for var, ret in zip(vars, v):
                        if var is not ret:
                            branch_ctx.add(f.lookup(builtins.eq, [var, ret]))
                    branches.append(branch_ctx.safe_wrap(vars))
            else:
                for branch in item._args:
                    branch_ctx = ctx.clone()
                    self.update(branch, branch_ctx)
                    branches.append(branch_ctx.safe_wrap([]))
            if isinstance(item, Union):
                ctx.add(f.union(branches, vars))
            else:
                ctx.add(f.match(branches, vars))

        elif isinstance(item, BranchRef):
            vars = ctx.value_map.get(item._match)
            if not vars:
                self.lookup(item._match, ctx)
                vars = ctx.value_map.get(item._match)
            assert isinstance(vars, list)
            return vars[item._ix]

        elif isinstance(item, Group):
            for g in item._group:
                self.lookup(g, ctx)

        elif isinstance(item, Distinct):
            vs = [self.lookup(v, ctx) for v in item._args]
            return flatten(vs)

        elif isinstance(item, (int, float, str)):
            return item

        elif item is None:
            return None

        elif hasattr(item, "_compile_lookup"):
            return item._compile_lookup(self, ctx)

        else:
            raise ValueError(f"Cannot lookup {item}")

    #--------------------------------------------------
    # Update
    #--------------------------------------------------

    def update(self, item:Expression|Match|Union, ctx:CompilerContext) -> ir.Value|list[ir.Var]:
        if isinstance(item, ConceptExpression):
            assert isinstance(item._op, Concept)
            relation = self.to_relation(item._op)
            (ident, kwargs) = item._params
            out = ctx.to_value(item)
            rels = {self.to_relation(getattr(item._op, k)): self.lookup(v, ctx)
                    for k, v in kwargs.items()}
            if ident is not None:
                out = self.lookup(ident, ctx)
                ctx.map_var(item, out)
            else:
                out = ctx.to_value(item._op)
                assert isinstance(out, ir.Var)
                ctx.add(f.construct(out, rels))
            assert not isinstance(out, list)
            ctx.add(f.derive(relation, [out]))
            for k, v in rels.items():
                assert not isinstance(v, list)
                ctx.add(f.derive(k, [out, v]))
            return out

        elif isinstance(item, Expression):
            params = flatten([self.lookup(p, ctx) for p in item._params])
            relation = self.to_relation(item._op)
            ctx.add(f.derive(relation, params))
            return params[-1]

        elif isinstance(item, Fragment):
            self.lookup(item, ctx)

        elif isinstance(item, (Match, Union)):
            self.lookup(item, ctx)

        elif hasattr(item, "_compile_update"):
            return item._compile_update(self, ctx)

        else:
            raise ValueError(f"Cannot update {item}")

__all__ = ["select", "where", "require", "then", "distinct", "per", "count", "sum", "min", "max", "avg"]

#--------------------------------------------------
# Todo
#--------------------------------------------------
"""
- Syntax
    ✔ construct
    ✔ static data handling
    ✔ Fix fragments to not be chained
    ✔ Extends
    ✔  Quantifiers
        ✔ not
        ✔ exists
        ✔ forall
    ✔ Aggregates
    ✔ Require
    ✘ Multi-step chaining
    ✔ ref
    ✔ alias
    ✔ match
    ✔ union
    ✔ capture all rules
    ✔ implement aliasing
    ✔ support defining relationships via madlibs Relationship("{Person} was born on {birthday:Date}")
    ✔ distinct
    ☐ nested fragments
    ☐  handle relationships with multiple name fields being accessed via prop:
        Package.shipment = Relationship("{Package} in {Shipment} on {Date}")
        Package.shipment.date, Package.shipment.shipment, Package.shipment.package
    ☐  sources
        ☐  table
        ☐  csv

- Compilation
    ✔ simple expressions
    ✔ select
    ✔ then
    ✔ Quantifiers
        ✔ exists
        ✔ forall
        ✔ not
    ✔ Aggregates
        ✔ Determine agg keys from inputs
        ✔ Group
    ✔ Require
    ✔ Alias
    ✔ Ref
    ✔ Match
    ✔ Union
    ✔ whole model
    ✔ distinct
    ✔ add Else to hoists
    ☐ extends
        ☐ nominals
    ☐ have require find keys and return the keys in the error
    ☐ Match/union with multiple branch refs in a select, duplicates the whole match
    ☐ nested fragments

☐ Execution
    ✔ basic queries
    ✔ query when iterating over a select
    ☐ debugger hookup
    ☐ table sources

"""



from __future__ import annotations
import re
import sys
from typing import Any, Literal
import warnings

from rich import terminal_theme as themes
from snowflake.snowpark import Session
from snowflake.snowpark.context import get_active_session

from ..clients.config import Config
from .base import NotebookCell, NotebookRuntimeEnvironment, SessionEnvironment, patch

class SnowbookEnvironment(NotebookRuntimeEnvironment, SessionEnvironment):
    remote = True
    last_run_id = None
    states = {}
    runner: Literal["container", "warehouse"] = "warehouse"

    def __init__(self):
        super().__init__()
        self.runner = "container" if "snowflake.connector.auth" in sys.modules else "warehouse"

    @classmethod
    def detect(cls):
        return "snowbook" in sys.modules

    def active_cell_id(self) -> str|None:
        return self.last_run_id

    def _patch(self):
        import snowbook # pyright: ignore

        @patch(snowbook.executor.notebook_compiler.NotebookCompiler, "compile_and_run_cell") # pyright: ignore
        def _(original, instance, cell, *args, **kwargs):
            id = self.last_run_id = cell.cell_content.id
            self.update_cell(NotebookCell(id, cell.cell_content.name, cell.cell_content.code, cell))
            state = original(instance, cell, *args, **kwargs)
            self.states[id] = state
            return state

        @patch(snowbook.executor.notebook_compiler, "execute_compilation") # pyright: ignore
        def _(original, *args, **kwargs):
            import streamlit as st # pyright: ignore
            try:
                return original(*args, **kwargs)
            except st.errors.UncaughtAppException as e:
                handled = self._handle_exc(e.exc)
                raise st.errors.UncaughtAppException(handled) if handled else e

        @patch(warnings, "showwarning")
        def _(original, message, category, filename, lineno, file=None, line=None):
            from ..errors import RAIWarning
            if isinstance(message, RAIWarning):
                self.report_alerts(message)
            else:
                original(message, category, filename, lineno, file, line)

    # Session Mgmt -------------------------------------------------------------

    def configure_session(self, config: Config, session: Any | None = None):
        if not session:
            session = get_active_session()

        if isinstance(session, Session):
            config.set("platform", "snowflake")
            user = session.sql("select current_user()").collect()[0][0]
            assert isinstance(user, str), "Could not retrieve current user"
            config.set("user", user)
            config.file_path = "__inline__"
            return session
        else:
            raise ValueError(f"Expected a snowpark Session object but received '{type(session)}'.")
    
    # Reporting ----------------------------------------------------------------

    def report_alerts(self, *alerts: Warning|BaseException):
       from ..errors import RAIException, RAIExceptionSet, RAIWarning, record

       import streamlit as st # pyright: ignore
       for alert in alerts:
          status = "warn" if isinstance(alert, Warning) else "error"
          if isinstance(alert, RAIExceptionSet):
             self.report_alerts(*alert.exceptions)
             continue

          elif isinstance(alert, RAIWarning) or isinstance(alert, RAIException):
             with record() as (console, _):
                alert.pprint()
                raw = console.export_html(inline_styles=True, code_format="{code}", theme=themes.NIGHT_OWLISH)
          else:
             raw = str(alert)

          formatted = self._format_alert(raw, status)
          st.markdown(formatted, unsafe_allow_html=True)

    def _format_alert(self, msg: str, status: Literal["error", "warn"] = "error"):
        for apply in [_format_alert_header, _format_alert_footer, _format_alert_line_breaks]:
           msg = apply(msg)

        return _format_alert_frame(msg, status)

    def _handle_exc(self, exc: BaseException):
        from ..errors import RAIException
        if isinstance(exc, RAIException):
           try:
              self.report_alerts(exc)
              return Exception("See above for details")
           except Exception as e:
              return e

#-------------------------------------------------------------------------------
# Report Formatting
#-------------------------------------------------------------------------------

FRAME_STYLE = {
    "warn": ("#fffce7", "#926c05"),
    "error": ("#ffecec", "#7d353b"),
}

def _format_alert_header(msg: str):
    def replace_func(match):
        span = match.group(1)
        title = match.group(2).strip()
        suffix = re.sub(r'</span>(?!.*<span)', '', match.group(3), flags=re.DOTALL) + (match.group(4) or "")
        rest = match.group(5) or ""
        return "".join([
            f'### <div style="display: flex">{span}{title}</span> ',
            f'<span style="flex: 1; text-align: right; font-size: 1rem;">{span}{suffix}</span>{rest}</span>'
        ])

    pattern = re.compile(r'(<span.*?>)--- (.*?) -+(.*?)(?:(: )|-+)</span>(.*)$', re.MULTILINE)
    return pattern.sub(replace_func, msg)

def _format_alert_footer(msg: str):
    pattern = re.compile(r'^(<p>\s*)?<span .*?>\s*----+\s*</span>(\s*</p>)?\s*$', re.MULTILINE)
    return pattern.sub("", msg)

def _format_alert_line_breaks(msg: str):
    return re.sub(r'(</\w+>)$', r'\1  ', msg, flags=re.MULTILINE)

def _format_alert_frame(msg: str, status: Literal["error", "warn"] = "error"):
    (bg, fg) = FRAME_STYLE[status]
    return "".join([
        f"""<div style="font-family: apercu-mono-regular, Menlo, Monaco, Consolas, 'Courier New', monospace;
                       padding: 8px 16px 0.1px 16px;
                       margin-bottom: 1em;
                       border-radius: 5px;
                       background-color: {bg};
                       color: {fg};">""".strip(),
       msg,
       "</div>"
    ])



import typing
from collections import defaultdict
from typing import TypeVar

from relationalai.early_access.dsl import Model, EntityType, ExternalRelation
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.types.standard import standard_value_types
from relationalai.early_access.dsl.core.utils import camel_to_snake
from relationalai.early_access.dsl.ontologies.constraints import Mandatory, Unique
from relationalai.early_access.dsl.ontologies.roles import AbstractRole
from relationalai.early_access.dsl.ontologies.subtyping import SubtypeConstraint, SubtypeArrow, \
    ExclusiveSubtypeConstraint, InclusiveSubtypeConstraint
from relationalai.early_access.metamodel.util import Printer

# Define a generic type variable constrained to SubtypeConstraint or its subclasses
T = TypeVar("T", bound=SubtypeConstraint)

class PythonPrinter(Printer):

    def to_python_string(self, model: Model) -> None:
        self._print_nl("import relationalai.early_access.dsl as rai")
        self._nl()

        self._print_nl(f"model = rai.Model(name='{model.name}', is_primary={model.is_primary})")
        self._print_nl("Concept = model.concept")
        self._print_nl("Entity = model.entity_type")
        self._print_nl("SubtypeArrow = model.subtype_arrow")
        self._print_nl("Relationship = model.relationship")
        self._print_nl("PrimaryKey = model.primary_key")
        self._print_nl("ExternalRelation = model.external_relation")

        self._nl()
        self._handle_value_types(model)

        self._nl()
        self._handle_entity_types(model)

        self._nl()
        self._handle_relationships(model)

        self._handle_composite_reference_schemas(model)

        self._nl()
        self._handle_external_relations(model)

        # TODO: handle Account.add_instance('Exchange') etc.

    def _handle_value_types(self, model: Model) -> None:
        for vt in model.value_types():
            self._print_nl(f"{vt.name()} = Concept('{vt.name()}', {', '.join(self._get_type(t) for t in vt._types)})")

    def _handle_entity_types(self, model: Model) -> None:
        subtype_arrows_by_type = self._group_subtype_arrows_by_type(model)
        inclusive_subtype_constraints_by_type = self._get_inclusive_subtype_constraints_by_type(model)
        exclusive_subtype_constraints_by_type = self._get_exclusive_subtype_constraints_by_type(model)

        for et in model.entity_types():
            name = et.name()
            if len(et.domain()) > 0:
                domain = ", ".join(self._get_type(t) for t in et.domain())
                self._print_nl(f"{name} = Entity('{name}', {domain})")
            else:
                self._print_nl(f"{name} = Entity('{name}')")
                self._handle_subtype_arrows(et,
                                            subtype_arrows_by_type.get(et, set()),
                                            inclusive_subtype_constraints_by_type.get(et, set()),
                                            exclusive_subtype_constraints_by_type.get(et, set()))

    def _handle_subtype_arrows(self, et: EntityType, subtype_arrows, inclusive_subtype_constraints,
                               exclusive_subtype_constraints) -> None:
        name = et.name()

        # Common elements in both sets
        common_constraints = inclusive_subtype_constraints & exclusive_subtype_constraints
        if len(common_constraints) > 0:
            for c in common_constraints:
                self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in c.arrows)}], "
                               f"exclusive=True, inclusive=True)")

        # Only in inclusive (but not in exclusive)
        only_inclusive_constraints = inclusive_subtype_constraints - exclusive_subtype_constraints
        if len(only_inclusive_constraints) > 0:
            for c in only_inclusive_constraints:
                self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in c.arrows)}], "
                               f"inclusive=True)")

        # Only in exclusive (but not in inclusive)
        only_exclusive_constraints = exclusive_subtype_constraints - inclusive_subtype_constraints
        if len(only_exclusive_constraints) > 0:
            for c in only_exclusive_constraints:
                self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in c.arrows)}], "
                               f"exclusive=True)")

        # Get all arrows from both inclusive and exclusive constraints
        all_arrows_in_constraints = {a for c in inclusive_subtype_constraints for a in c.arrows} | \
                                    {a for c in exclusive_subtype_constraints for a in c.arrows}

        # Subtract the arrows in constraints from subtype_arrows
        remaining_constraints = subtype_arrows - all_arrows_in_constraints
        if len(remaining_constraints) > 0:
            for c in remaining_constraints:
                self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in c.arrows)}])")

    def _handle_relationships(self, model: Model) -> None:
        unique_roles = self._get_unique_roles(model)
        mandatory_roles = self._get_mandatory_roles(model)
        internal_preferred_identifier_roles = self._get_internal_preferred_identifier_roles(model)

        for rel in model.relationships():
            if not rel.is_subtype() and not rel.is_identifier():
                name = camel_to_snake(rel._name())
                self._print_nl(f"{name} = Relationship()")
                self._print_nl(f"with {name} as rel:")
                self._handle_relationship_roles(rel, mandatory_roles, unique_roles, internal_preferred_identifier_roles)
                self._handle_relationship_relations(rel)
                self._nl()

    def _handle_relationship_roles(self, rel, mandatory_roles, unique_roles, internal_preferred_identifier_roles):
        for i, r in enumerate(rel.roles()):
            self._indent_print_nl(1, f"rel.role({self._get_type(r.player())}"
                                     f"{self._print_if_not_empty('name', r.name())}"
                                     f"{self._print_if_true('unique', r in unique_roles)}"
                                     f"{self._print_if_true('mandatory', r in mandatory_roles)}"
                                     f"{self._print_if_true('primary_key', r in internal_preferred_identifier_roles)})")
            if r.prefix or r.postfix:
                elements = list(filter(None, [
                    self._print_first_if_not_empty('prefix', r.prefix),
                    self._print_first_if_not_empty('postfix', r.postfix)
                ]))
                self._indent_print_nl(1, f"rel.role_at({i}).verbalization({', '.join(elements)})")

    def _handle_relationship_relations(self, rel):
        for relation in rel.relations():
            if relation.reading():
                reading = relation.reading()

                num_roles = len(rel.roles())
                num_texts = len(reading.text_frags)

                elements = []

                for i in range(num_texts):
                    elements.append(f"rel.role_at({i})")  # Role first
                    elements.append(f"'{reading.text_frags[i]}'")  # Then text

                if num_roles > num_texts:
                    elements.append(f"rel.role_at({num_roles - 1})")  # Add last role if needed

                self._indent_print_nl(1, f"rel.relation({', '.join(elements)}"
                                         f"{self._print_if_not_empty('name', reading.rel_name)}"
                                         f"{self._print_if_true('functional', relation.signature().functional())})")

    def _handle_composite_reference_schemas(self, model: Model) -> None:
        for pi in self._get_composite_preferred_identifiers(model):
            elements = [f"{camel_to_snake(r.part_of._name())}.role_at(1)" for r in pi.roles()]
            self._print_nl(f"PrimaryKey({', '.join(elements)})")

    def _handle_external_relations(self, model: Model) -> None:
        external_relations = self._get_external_relations(model)
        for r in external_relations:
            self._print_nl(f"ExternalRelation('{r.rel_name()}', {', '.join(self._get_type(t) for t in r.signature().types())})")

    @staticmethod
    def _get_mandatory_roles(model: Model) -> set[AbstractRole]:
        return {c.role for c in model.constraints() if isinstance(c, Mandatory)}

    @staticmethod
    def _get_unique_roles(model: Model) -> set[AbstractRole]:
        return {role for c in model.constraints() if isinstance(c, Unique) and not c.is_preferred_identifier for role in c.roles()}

    @staticmethod
    def _get_internal_preferred_identifier_roles(model: Model) -> set[AbstractRole]:
        return {role for c in model.constraints() if isinstance(c, Unique) and c.is_preferred_identifier and len(c.roles()) == 1 for role in c.roles()}

    @staticmethod
    def _get_composite_preferred_identifiers(model: Model) -> set[Unique]:
        return {c for c in model.constraints() if isinstance(c, Unique) and c.is_preferred_identifier and len(c.roles()) > 1}

    @staticmethod
    def _get_type(t: Type) -> str:
        return f"rai.{t.name()}" if t.name() in standard_value_types else t.name()

    @staticmethod
    def _group_subtype_arrows_by_type(model: Model) -> dict[Type, set[SubtypeArrow]]:
        subtype_arrows_by_type = defaultdict(set)
        for a in model.subtype_arrows():
            subtype_arrows_by_type[a.end].add(a)
        return dict(subtype_arrows_by_type)

    @staticmethod
    def _get_subtype_constraints_by_type(model: Model, constraint_type: typing.Type[T]) -> dict[Type, set[T]]:
        constraints_by_type: dict[Type, set[T]] = defaultdict(set)

        for c in model.subtype_constraints():
            if isinstance(c, constraint_type):
                for a in c.arrows:
                    constraints_by_type[a.end].add(c)

        return dict(constraints_by_type)

    def _get_exclusive_subtype_constraints_by_type(self, model: Model) -> dict[Type, set[ExclusiveSubtypeConstraint]]:
        return self._get_subtype_constraints_by_type(model, ExclusiveSubtypeConstraint)

    def _get_inclusive_subtype_constraints_by_type(self, model: Model) -> dict[Type, set[InclusiveSubtypeConstraint]]:
        return self._get_subtype_constraints_by_type(model, InclusiveSubtypeConstraint)

    @staticmethod
    def _get_external_relations(model: Model) -> set[ExternalRelation]:
        return {r for r in model.relations() if isinstance(r, ExternalRelation)}

    @staticmethod
    def _print_if_not_empty(label: str, value: str) -> str:
        return f", {label}='{value}'" if value else ""

    @staticmethod
    def _print_first_if_not_empty(label: str, value: str) -> str:
        return f"{label}='{value}'" if value else ""

    @staticmethod
    def _print_if_true(label: str, value: bool) -> str:
        return f", {label}=True" if value else ""


from relationalai.early_access.dsl.core import warn
from relationalai.early_access.dsl.core.builders.logic import LogicBuilder
from relationalai.early_access.dsl.core.cset import ConstraintSet
from relationalai.early_access.dsl.core.exprs.scalar import ScalarVariable, ScalarExprBuilder

class RelVariable(ScalarVariable):

    def __init__(self, nm, tp):
        self._name = nm
        self._type = tp

    def display(self): return self._name

    def entityid(self):
        return hash((self._name, self._type.entityid()))

    def physical_typeof(self):
        return self.typeof().physical_typeof()

    def typeof(self): return self._type

    def __hash__(self):
        return hash((self._name, self._type))

class LogicFragment(ConstraintSet, LogicBuilder, ScalarExprBuilder):

    def __init__(self):
        ConstraintSet.__init__(self)
        self._scalars = {}
        self._existentials = {}

    def add_component(self, d):
        vname = d.display()

        if vname not in self._scalars:
            self._scalars[vname] = d
            return True
        else:
            d1 = d.typeof()
            d2 = self._scalars[vname].typeof()
            if not d2.subtype_of(d1):
                if d1.subtype_of(d2):
                    self._scalars[vname] = d
                else:
                    raise Exception(
                        f"++ Error: Cannot declare two scalar variables named {vname} in the same schema context")

        return False

    def add_existential_constraint(self, c):
        self._existentials[id(c)] = c
        return c

    def build_atom(self, rel, args):
        atom = LogicBuilder.build_atom(self, rel, args)
        self.add_atomic_constraint(atom)
        return atom

    def build_comparison(self, left, op, right):
        c = LogicBuilder.build_comparison(self, left, op, right)
        self.add_scalar_constraint(c)
        return c

    def build_scalar_variable(self, args, kwargs):
        tp = args[0]
        if len(args) < 2 or not isinstance(args[1], str):
            raise Exception("This should not happen")

        decl = RelVariable(args[1], tp)
        self.add_component(decl)
        return decl

    def build_relation_variable(self, args, kwargs):
        tp = args[0]

        # [REKS] Very subtle. This is needed because when tp is a ConstrainedValueType,
        #        this method will be called. Correct during type analysis, but incorrect
        #        during declarations of Rel variables of said type
        #
        if tp.nominal():
            return self.build_scalar_variable(args, kwargs)

        # [REKS] Another subtlety. In Rel, we can name a Relation the same as an EntityType,
        #        and we often do this for relations that maintain the finite population of
        #        entites of that type. But then we can refer to these relations as Atoms in
        #        rules, and they will be interpreted as requests to instantiate the EntityType
        #        with a variable name -- which is not a name but a RelVariable meant to range
        #        over the population of the relation.
        #
        varorname = args[1]
        if isinstance(varorname, RelVariable):
            # [VAMI] TODO: uncomment later
            # warn(f"Corner case to follow up on when building_relation_varornameiable {varorname.display()}")
            return self.build_atom(tp.population(), [varorname])
        else:
            decl = RelVariable(varorname, tp)

            self.add_component(decl)
            return decl


    # Partially evaluate this Schema by substituting values for
    # its schema variables, simplifying where possible.
    #
    def evaluate(self, bdgs):

        # Project away from decls any variables that are declared in pdecls
        #
        def project_decls(decls, pdecls):
            return {v: decls[v] for v in decls if v not in pdecls}

        decls = project_decls(self._scalars, bdgs)

        super = ConstraintSet.evaluate(self, bdgs)

        if super is None:
            return super
        else:
            s = LogicFragment()
            s._scalars = decls
            s._sconstraints = super._sconstraints
            s._atoms = super._atoms
            return s

    def propositional_constraints(self):
        par = ConstraintSet.propositional_constraints(self)
        if len(self._existentials) == 0:
            return par
        else:
            d = self._existentials
            for x in d:
                par[x] = d[x]
        return par

    def rel_formula(self):
        subf = ConstraintSet.rel_formula(self)
        if len(self._existentials) == 0:
            if len(subf) == 0:
                return '    true'
            else:
                return subf
        else:
            xcs = [c.rel_formula() for c in self._existentials.values()]
            xf = "\n".join(xcs)
            if subf == "":
                return xf
            return subf + "\n" + xf

    # Given an optional set of constraints that defaults to all propositional
    # constraints in this LogicFragment, returns a dictionary that represents
    # that subset of this Fragment's scalar variables that is not grounded by
    # those constraints.
    #
    # Grounding is decided the same way it is in datalog, namely if the
    # variable is referenced positively in a finite relation or in an
    # atom of an infinite relation in which every other variable in that
    # atom is grounded.
    #
    def ungrounded_components(self, constraints=None):

        if constraints is None:
            constraints = self.propositional_constraints()
        vars = [d for d in self._scalars.values()]

        groundings = {}
        glen = -1

        while glen < len(groundings):
            glen = len(groundings)

            for c in constraints.values():
                for v in vars:
                    if c.grounds(v, groundings):
                        groundings[v.display()] = v

        return [v for v in vars if v.display() not in groundings]

    def view(self, view_name):
        c = self.ungrounded_components()

        args = ", ".join([v.display() for v in self._scalars.values()])
        body = self.rel_formula()
        rule = f"def {view_name}({args}): " + "\n" + body

        if len(c) > 0:
            vstr = ", ".join([v.display() for v in c])
            warn(f"Ungrounded variables ({vstr}) in schema prevent materialization (declaring view @inline)")
            return "@inline\n" + rule
        else:
            return rule



from relationalai.early_access.dsl.core.exprs import Expr


class Instance(Expr):

    def __init__(self, c, arglist):
        self.concept = c
        self.arglist = arglist

    def entityid(self):
        arghashes = []
        for c in self.arglist:
            if isinstance(c, Instance):
                arghashes.append(c.entityid())
            else:
                arghashes.append(c)
        return hash((self.concept.entityid(), *arghashes))

    def display(self):
        params = []
        for i in range(len(self.arglist)):
            arg = self.arglist[i]
            if isinstance(arg, Instance):
                params.append(arg.pprint())
            else:
                # Otherwise, arg must be a constraint of some UnconstrainedValueType
                if isinstance(arg, str):
                    params.append(f"\"{arg}\"")
                else:
                    params.append(arg)
        return self.concept.ctor_name() + '[' + ",".join(params) + ']'

    def pprint(self):
        return self.display()


class ValueTypeInstance(Instance):

    def __init__(self, vtype, arglist):
        if len(arglist) != len(vtype._params):
            raise Exception(
                f"Value type {vtype.name()} constructor maps tuples of length {len(vtype.constructor())} but was "
                f"instantiated with {len(arglist)} arguments")
        Instance.__init__(self, vtype, arglist)



from relationalai.early_access.dsl.core.constraints.scalar import fold_constants, propagate_definitions, entails, \
    unsatisfiable
from relationalai.early_access.dsl.core.types.unconstrained import UnconstrainedValueType


# Counts the number of constraints of props in which variable x occurs.
#
def occurrences(x: str, props):
    count = 0
    for c in props.values():
        if c.refersto(x):
            count += 1
    return count


# Given a LogicFragment of the form:
#
#     (bvars: sconsts & atoms)
#
# where:
#
#   - bvars is a set of (Scalar) bound variable declarations,
#   - sconsts is a dictionary of ScalarConstraints, and
#   - atoms is a dictionary of AtomicConstraints
#
# Note that constraints in sconsts and/or atoms could refer to *free variables*,
# i.e., ScalarVariables that are not in bvars. We use type constraints on free
# variables when reasoning to help check domain constraints on bound variables.
#
# This function folds constraints and applies the one-point rule where possible to
# generate a hopefully simpler but equivalent fragment (bvars' : sconsts' & atoms')
# where:
#
#    - |bvars'| <= |bvars|
#    - |sconsts'| <= |sconsts|
#
# Recall the formal definition of the one-point rule:
#
#       exists(x in D : P and (x=t) )
#      =============================== [one-point]
#             t in D and P[t/x]
#
# To apply it, we take advantage of the fact that constant folding will often do
# most of the substitution work (i.e., P[t/x]) for ScalarConstraints of the form:
#
#     x == t
#
# which means we can apply the rule by merely removing such constraints (and in
# some cases also the variable ScalarVariable x) provided that we can stacially
# check that t is in D, where D is the (value-) type of ScalarVariable x. This
# check is called the "domain check" on t. Note that we cannot always statically
# discharge the domain check.
#
def one_point_reduce(bvars, sconsts, atoms):

    if len(sconsts) == 0:
        return bvars, sconsts, atoms

    # Start by folding constants
    #
    (sc_folded, atoms_folded) = fold_constants(sconsts, atoms)

    # Next infer the set of free variables (those referenced in some constraint but
    # that are not in bvars) so that we can commute and then and propagate definitions.
    #
    freevars = free_variable_refs(bvars, sc_folded, atoms_folded)

    # Now that we know the free variables, go instantiate their type constraints so
    # that we can leverage them to do more precise domain checking below.
    #
    free_var_constraints = {}
    for v in freevars.values():
        free_var_constraints = free_var_constraints | instantiate_type_constraints_for(v, v._scalartype)

    sc_commuted = commute_definitions(freevars, sc_folded)

    # Next, propagate definitions
    #
    (sc_propagated, atoms_propagated) = propagate_definitions(sc_commuted, atoms_folded)

    # Post: Any ScalarVariable `x` in bvars that forms the lhs of some definition constraint
    #       (of the form "x == t") can now be considered for elimination via the one-point
    #       rule.
    #
    # We choose to eliminate the constraint (and the variable 'x') if 'x' occurs only in that
    # constraint and if we can prove the domain check on `x` is satisfied
    #

    # Here are the variables from definition constraints that occur only in that definition
    # constraint.
    #
    defns = {c.left.display(): c for c in sc_propagated.values() if c.definition()}
    occurs_once = [x for x in bvars
                   if x in defns and occurrences(x, sc_propagated) == 1 and occurrences(x, atoms_propagated) == 0]

    # Here is the accumulated knowledge we can appeal to when verifying domain checks to try to remove
    # one of these variables
    #
    domain_knowledge = sc_propagated | free_var_constraints

    # Any bound that occurs once and satisfies the domain constraint can be removed.
    #
    vars_to_remove = [x for x in occurs_once if check_domain_constraints(bvars[x], domain_knowledge)]

    # Compute the set of new declarations to place under an existential quantifier.
    # These should include any bvars that we have *not* selected for removal by
    # virtue of applying the one-point rule.
    #
    existential_vars = {v: bvars[v] for v in bvars if v not in vars_to_remove}

    # These constraints trigger an application of the one-point rule and so can
    # be removed.
    #
    triggers = [defns[x] for x in vars_to_remove]

    # These are the constraints that remain after removing those that were
    # selected for removal using the one-point rule
    #
    sconsts_prime = {c.entityid(): c for c in sc_propagated.values() if c not in triggers}

    return existential_vars, sconsts_prime, atoms_propagated


def free_variable_refs(bvars, sconsts, atoms):
    fvars = {}
    for c in sconsts.values():
        refs = c.scalar_refs()
        for r in refs:
            if r not in bvars:
                fvars[r] = refs[r]

    for a in atoms.values():
        refs = a.scalar_refs()
        for r in refs:
            if r not in bvars:
                fvars[r] = refs[r]

    return fvars

# We commute any commutative definitions that involve a free variable
# to try to ensure the bound variable is in the lhs of the constratint
#
def commute_definitions(fvars, constraints):
    commuted_constraints = {}
    for c in constraints:
        const = constraints[c]
        if const.commutative_definition() and const.left.display() in fvars:

            newc = const.commute()
            commuted_constraints[newc.entityid()] = newc
        else:
            commuted_constraints[c] = const

    return commuted_constraints


# Given a ScalarVariable var and a ConstrainedValueType tp,
# returns the set of type constraints on tp where the type
# variable is replaced by var -- essentially "instantiating"
# those constraints on var.
#
def instantiate_type_constraints_for(var, tp):
    constraints = {}
    if tp.constraint_unary():
        tvar = tp.value()
        typecs = tp.verifies()
        for c in typecs.values():
            xc = c.substitute({tvar.display(): var})
            constraints[xc.entityid()] = xc
    return constraints

# Let v be a ScalarComponent and comps be a set of ScalarConstraints.
# Returns true if comps entails all of the type constraints of v.
# We use this function to check the domain constraint of v when
# applying the one-point rule.
#
def check_domain_constraints(v, comps):
    if isinstance(v, UnconstrainedValueType):
        return not unsatisfiable(comps)
    else:
        typecs = instantiate_type_constraints_for(v, v._scalartype)
        for c in typecs.values():
            if not entails(comps, c):
                return False

    return True


from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
import json
import re
import textwrap
from snowflake.snowpark.context import get_active_session
from . import builder as b

#--------------------------------------------------
# Source
#--------------------------------------------------

class Source(b.Producer):
    def __init__(self) -> None:
        super().__init__(None)
        self._relationships = {}

#--------------------------------------------------
# Snowflake sources
#--------------------------------------------------

SFTypes = {
    "TEXT": "String",
    "FIXED": "Number",
}

def parse_fqn(fqn:str):
    database, schema, table = fqn.split(".")
    return database, schema, table

SF_ID_REGEX = re.compile(r'^[A-Za-z_][A-Za-z0-9_$]*$')
def quoted(ident:str):
    if SF_ID_REGEX.match(ident) or ident[0] == '"':
        return ident
    return f'"{ident}"'

@dataclass
class TableInfo:
    source:SnowflakeTable|None
    fields:list[b.Field]
    raw_columns:list[dict]

class SchemaInfo:
    def __init__(self, database:str, schema:str) -> None:
        self.database = database
        self.schema = schema
        self.tables = defaultdict(lambda: TableInfo(None, [], []))
        self.fetched = False

    def fetch(self):
        self.fetched = True
        session = get_active_session()
        table_names = self.tables.keys()
        name_lookup = {x.upper(): x for x in table_names}
        tables = ", ".join([f"'{x.upper()}', '{x}'" for x in self.tables.keys()])
        columns = session.sql(textwrap.dedent(f"""
            begin
                SHOW COLUMNS IN SCHEMA {quoted(self.database)}.{quoted(self.schema)};
                let r resultset := (select "table_name", "column_name", "data_type" from table(result_scan(-1)) as t
                                    where "table_name" in ({tables}));
                return table(r);
            end;
        """))
        print(textwrap.dedent(f"""
            begin
                SHOW COLUMNS IN SCHEMA {quoted(self.database)}.{quoted(self.schema)};
                let r resultset := (select "table_name", "column_name", "data_type" from table(result_scan(-1)) as t
                                    where "table_name" in ({tables}));
                return table(r);
            end;
        """))
        columns = columns.collect()
        print(columns)
        for row in columns:
            table_name, column_name, data_type = row
            table_name = name_lookup.get(table_name, table_name)
            info = self.tables[table_name]
            type_str = SFTypes[json.loads(data_type).get("type")]
            info.fields.append(b.Field(name=column_name, type_str=type_str, is_many=False))
            info.raw_columns.append(row.as_dict())

class SnowflakeTable(Source):
    _schemas:dict[tuple[str, str], SchemaInfo] = {}

    def __init__(self, fqn:str) -> None:
        super().__init__()
        self._fqn = fqn
        self._database, self._schema, self._table = parse_fqn(fqn)
        self._inited = False
        info = self._schemas.get((self._database, self._schema))
        if not info:
            info = self._schemas[(self._database, self._schema)] = SchemaInfo(self._database, self._schema)
        info.tables[self._table].source = self

    def _lazy_init(self):
        if self._inited:
            return
        self._inited = True
        schema_info = self._schemas[(self._database, self._schema)]
        if not schema_info.fetched:
            schema_info.fetch()
        table_info = schema_info.tables[self._table]
        print(table_info.raw_columns)
        self._rel = b.Relationship(self._fqn, fields=[b.Field("row_id", "Number", False)] + table_info.fields)
        print(self._rel, self._rel._fields)



    def _compile_lookup(self, compiler:b.Compiler, ctx:b.CompilerContext):
        self._lazy_init()
        return


#--------------------------------------------------
# Convenience functions
#--------------------------------------------------

def table(name:str) -> SnowflakeTable:
    return SnowflakeTable(name)



# Stack abstraction
from typing import Optional


class Node:
    def __init__(self, value):
        self.value = value
        self.next: Optional[Node] = None


class Stack:

    def __init__(self):
        self._head = None
        self._size = 0

    # Get the current size of the stack
    def size(self):
        return self._size

    # Check if the stack is empty
    def empty(self):
        return self._size == 0

    def top(self):
        if self.empty() or self._head is None:
            return None

        return self._head.value

    def push(self, value):
        node = Node(value)
        node.next = self._head
        self._head = node
        self._size += 1

    def pop(self):
        if self.empty() or self._head is None:
            raise Exception("Popping from an empty stack")
        top_element = self._head
        self._head = top_element.next
        self._size -= 1
        return top_element.value
